﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.keytable;

// キーテーブル SQLiteバージョン

private import std.string;
private import std.utf;
private import std.stdio;
private import std.random;

private import pes.wstring;
private import pes.conv;
private import pes.random;
private import pes.net.winny.key;
private import pes.system.delay;
private import pes.system.semaphore;
private import pes.data.sqlite3;
private import pes.system.task;

import pes.prompt;

// キー要素
class KeyInformation
{
protected:
	NyKeyInformation	m_key;
	
public:
	NyKeyInformation key()
	{
		return m_key;
	}
	
	// ソート用
	override int opCmp(Object o)
	{
		KeyInformation k = cast(KeyInformation)o;
		
		return k.m_key.timer - m_key.timer;
	}
	
	this(NyKeyInformation k)
	{
		m_key 		= k;
	}
	
	ubyte[] aaKey()
	{
		return m_key.hash;
	}
}

/* SQLなど	*/

// ここに書くのは汚いなー･･･

const char[] KEYTABLE_CREATE = 
//	"CREATE TEMP TABLE KEY_TABLE(hash blob, size int, name text, trip text, timer int, packet blob)";
"CREATE TEMP TABLE KEY_TABLE(
	hash blob, size int, name text, trip text, timer int, 
	mbs_length int, key_address int, key_port int, bbs_address int, bbs_port int, 
	bbs_trip blob, ref_block int, mtime int, ignore int, key_version int)";


const char[] KEYTABLE_HASH_INDEX_CREATE = 
	"CREATE UNIQUE INDEX KEY_TABLE_HASH_IDX on KEY_TABLE(hash)";

const char[] KEYTABLE_NAME_INDEX_CREATE = 
	"CREATE INDEX KEY_TABLE_NAME_IDX on KEY_TABLE(name)";

const char[] KEYTABLE_TTL_INDEX_CREATE = 
	"CREATE INDEX KEY_TABLE_TIMER_IDX on KEY_TABLE(timer)";
	

const char[] ADDRESSTABLE_CREATE = 
"CREATE TEMP TABLE ADDRESS_TABLE(hash blob, size int, address int, port address)";

const char[] KEYTABLE_DROP =
	"DROP TABLE KEY_TABLE";
	
const char[] KEYTABLE_CLEAR =
	"DELETE FROM KEY_TABLE";

const char[] KEYTABLE_CLEAR_COUNT =
"DELETE FROM KEY_TABLE WHERE hash IN (SELECT hash FROM KEY_TABLE ORDER BY timer ASC LIMIT ?)";

const char[] KEYTABLE_INSERT =
	"INSERT INTO KEY_TABLE 
(
	hash , size , name , trip , timer , 
	mbs_length , key_address , key_port , bbs_address , bbs_port , 
	bbs_trip, ref_block , mtime , ignore , key_version 
)
values (
	?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?
);";

const char[] KEYTABLE_EXISTS_DELETE =
	"DELETE FROM KEY_TABLE WHERE hash=?";

const char[] KEYTABLE_SERACH_HASH =
	"SELECT * FROM KEY_TABLE WHERE hash=? LIMIT ?";

const char[] KEYTABLE_SERACH_NAME =
	"SELECT * FROM KEY_TABLE WHERE name like('%'|| ? ||'%') LIMIT ?";

const char[] KEYTABLE_SERACH_NAME_RANDOM =
	"SELECT * FROM KEY_TABLE WHERE name like('%'|| ? ||'%') order by mtime % ? LIMIT ?";

const char[] KEYTABLE_SERACH_TRIP =
	"SELECT * FROM KEY_TABLE WHERE trip = ? LIMIT ?";

const char[] KEYTABLE_TTL_DOWN = 
	"UPDATE KEY_TABLE SET timer = timer - 1";

const char[] KEYTABLE_TTL_DOWN30 = 
	"UPDATE KEY_TABLE SET timer = timer - 30";

const char[] KEYTABLE_TTL_DIE = 
	"DELETE FROM KEY_TABLE WHERE timer <= 0";

const char[] KEYTABLE_COUNT =
	"SELECT COUNT(*) FROM KEY_TABLE";

const char[] KEYTABLE_DB_FILE = "keytable.db";


// テスト用
extern (C) int callback_func(void *NotUsed, int argc, char **argv, char **azColName){
  int i;
  for(i=0; i<argc; i++){
    printf("%s = %s\n", azColName[i], argv[i] ? argv[i] : "NULL");
  }
  printf("\n");
  return 0;
}

private class SearchQuery
{
protected:
	char[] 	m_query;
	bool	m_first;
	
public:
	this()
	{
		m_query = "SELECT * FROM key_table WHERE ";
		m_first = true;
	}
	
	void and(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		
		m_query ~= add ~ std.string.format("name like ('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	void not(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		m_query ~= add ~ std.string.format("name not like('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	char[] text()
	{
		return m_query ~ " LIMIT ?";
	}
	
	char[] textRandom()
	{
		return m_query ~ " ORDER BY size % ? LIMIT ?";
	}
	
	void reset()
	{
		m_query = null;
	}
}


private class CountQuery
{
protected:
	char[] 	m_query;
	bool	m_first;
	
public:
	this()
	{
		m_query = "SELECT count(*) FROM key_table WHERE ";
		m_first = true;
	}
	
	void and(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		
		m_query ~= add ~ std.string.format("name like ('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	void not(char[] keyword)
	{
		char[] add;
		
		if (! m_first) {
			add = " AND ";
		} else {
			m_first = false;
		}
		m_query ~= add ~ std.string.format("name not like('%%%s%%')",
												escapeSQLText(keyword));
	}
	
	char[] text()
	{
		return m_query ~ " LIMIT ? ";
	}
	
	char[] textRandom()
	{
		return m_query.dup ~ " LIMIT ? ";
	}
	
	void reset()
	{
		m_query = null;
	}
}


private char[] makeQuery(char[] input)
{
	char[][] scrap = std.string.split(input);
	SearchQuery query = new SearchQuery;
	
	foreach (char[] part; scrap) {
		if (part.length == 0 ) {
			continue;
		}
		if (part[0] == '-') {
			query.not(part[1 .. length]);
		} else {
			query.and(part);
		}
	}
	
	return query.text;
}
private char[] makeCountQuery(char[] input)
{
	char[][] scrap = std.string.split(input);
	CountQuery query = new CountQuery;
	
	foreach (char[] part; scrap) {
		if (part.length == 0 ) {
			continue;
		}
		if (part[0] == '-') {
			query.not(part[1 .. length]);
		} else {
			query.and(part);
		}
	}
	
	return query.text;
}


private char[] makeQueryRandom(char[] input)
{
	char[][] scrap = std.string.split(input);
	SearchQuery query = new SearchQuery;
	
	foreach (char[] part; scrap) {
		if (part.length == 0 ) {
			continue;
		}
		if (part[0] == '-') {
			query.not(part[1 .. length]);
		} else {
			query.and(part);
		}
	}
	
	return query.textRandom;
}

private char[] makeCountQueryRandom(char[] input)
{
	char[][] scrap = std.string.split(input);
	CountQuery query = new CountQuery;
	
	foreach (char[] part; scrap) {
		if (part.length == 0 ) {
			continue;
		}
		if (part[0] == '-') {
			query.not(part[1 .. length]);
		} else {
			query.and(part);
		}
	}
	
	return query.textRandom;
}

bool nullQuery(char[] s)
{
	foreach (char c; s) {
		if (c == 0) {
			return true;
		}
	}
	
	foreach (char c; s) {
		if (c != ' ' && c != '-') {
			return false;
		}
	}
	
	return true;
}


// キーテーブル
class KeyTable
{
protected:
	SQLite 				m_db;
	SemaphoreResource 	m_semaphore;
	semaphore			m_id;
	
public:
	void signal()
	{
		m_semaphore.signal(m_id);
	}
	
	void release()
	{
		m_semaphore.release(m_id);
	}
	
	this(SemaphoreResource resource, semaphore id)
	{
		m_semaphore	= resource;
		m_id		= id;
		open();
		
		try {
			m_db.exec(KEYTABLE_DROP);
		} catch (SQLiteException) {
			//
		}
		m_db.exec(KEYTABLE_CREATE);
		m_db.exec(KEYTABLE_HASH_INDEX_CREATE);
		m_db.exec(KEYTABLE_TTL_INDEX_CREATE);
		m_db.exec(KEYTABLE_NAME_INDEX_CREATE);
	}
	
	void open()
	{
		signal();
		try {
			if (m_db !is null) {
				close();
			}
			m_db 		= new SQLite(KEYTABLE_DB_FILE);
		} finally {
			release();
		}
	}
	
	void close()
	{
		signal();
		try {
			m_db.close();
			m_db = null;
		} finally {
			release();
		}
	}
	
	void reopen()
	{
		signal();
		try {
			close();
			open();
		} finally {
			release();
		}
	}
	
	void addKeyArray(NyKeyInformation[] array)
	{
		KeyInformation key;
		
		signal();
		m_db.begin();
		try {
			
			auto SQLiteStatement insert_stmt = m_db.prepare(KEYTABLE_INSERT);
			auto SQLiteStatement delete_stmt = m_db.prepare(KEYTABLE_EXISTS_DELETE);
			
			foreach (NyKeyInformation e; array) {
				delete_stmt.bind(1, e.hash);
				
				insert_stmt.bind(1, e.hash, e.fileSize, e.fileName, e.trip, e.timer, 
									e.fileNameMBSLength, e.keyAddress, e.keyPort, 
									e.bbsAddress, e.bbsPort, e.bbsTrip, e.referredBlock, 
									e.modifiedTime, e.ignore, e.keyVersion);
				
				delete_stmt.step();
				insert_stmt.step();
				
				delete_stmt.reset();
				insert_stmt.reset();
			}
		} catch (Exception e) {
			log("Exception: addKeyArray");
			throw e;
		} finally {
//			delete_stmt.finalize();
//			insert_stmt.finalize();
			m_db.commit();
			release();
		}
	}
	
	NyKeyInformation[] searchHash(ubyte[] hash, uint max = uint.max)
	{
		NyKeyInformation[] result;
		uint i, pos;
		
		signal();
		m_db.begin();
		try {
			auto SQLiteStatement stmt = m_db.prepare(KEYTABLE_SERACH_HASH);
			
			stmt.bind(1, hash, max);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {// 本当は回るわけない
				NyKeyInformation key 	= new NyKeyInformation;
				pos = 0;
				
				key.hash				= stmt.fetch!(ubyte[])(pos++).dup;
				key.fileSize			= stmt.fetch!(uint)(pos++);
				key.fileName			= stmt.fetch!(char[])(pos++).dup;
				key.trip				= stmt.fetch!(char[])(pos++).dup;
				key.timer				= stmt.fetch!(uint)(pos++);
				key.fileNameMBSLength	= stmt.fetch!(uint)(pos++);
				key.keyAddress 			= stmt.fetch!(uint)(pos++);
				key.keyPort				= stmt.fetch!(ushort)(pos++);
				key.bbsAddress			= stmt.fetch!(uint)(pos++);
				key.bbsPort				= stmt.fetch!(ushort)(pos++);
				key.bbsTrip				= stmt.fetch!(ubyte[])(pos++);
				key.referredBlock		= stmt.fetch!(uint)(pos++);
				key.modifiedTime		= stmt.fetch!(uint)(pos++);
				key.ignore				= stmt.fetch!(ubyte)(pos++);
				key.keyVersion			= stmt.fetch!(ubyte)(pos++);
				
				if (key.timer <= 0) {
					continue;
				}
				result ~= key;
				if (++i >= max) {
					break;
				}
			}
		} catch (Exception e) {
			log("Exception: searchHash");
			throw e;
		} finally {
//			stmt.finalize();
			m_db.end();
			release();
		}
		
		return result;
	}
	
	NyKeyInformation[] searchName(char[] keyword, uint max = uint.max)
	{
		NyKeyInformation[] result;
		uint i, pos;
		
		if (keyword.length == 0 || nullQuery(keyword)) {
			return null;
		}
		if (keyword[0] == '%' && (keyword.length -1) % 2 == 0) {
			try {
				return searchHash(hexStringToBinary(keyword[1 .. length]), max);
			} catch (ConvError) {
				return null;
			}
			
			return null;
		}
		char[] query 		 = makeQuery(keyword);
		char[] qcount		 = makeCountQuery(keyword);
		
		signal();
		m_db.begin();
		try {
			
			uint count;
			auto SQLiteStatement count_stmt = m_db.prepare(qcount);
			count_stmt.bind(1, max);
			count_stmt.step();
			count = count_stmt.fetch!(uint)(0);
			
			if (count > max) {
				count = max;
			}
			
			result = new NyKeyInformation[count];
			
			auto SQLiteStatement stmt = m_db.prepare(query);
			
			stmt.bind(1, max);
			
			for (i = 0, stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				pos = 0;
				
				key.hash				= stmt.fetch!(ubyte[])(pos++).dup;
				key.fileSize			= stmt.fetch!(uint)(pos++);
				key.fileName			= stmt.fetch!(char[])(pos++).dup;
				key.trip				= stmt.fetch!(char[])(pos++).dup;
				key.timer				= stmt.fetch!(uint)(pos++);
				key.fileNameMBSLength	= stmt.fetch!(uint)(pos++);
				key.keyAddress 			= stmt.fetch!(uint)(pos++);
				key.keyPort				= stmt.fetch!(ushort)(pos++);
				key.bbsAddress			= stmt.fetch!(uint)(pos++);
				key.bbsPort				= stmt.fetch!(ushort)(pos++);
				key.bbsTrip				= stmt.fetch!(ubyte[])(pos++).dup;
				key.referredBlock		= stmt.fetch!(uint)(pos++);
				key.modifiedTime		= stmt.fetch!(uint)(pos++);
				key.ignore				= stmt.fetch!(ubyte)(pos++);
				key.keyVersion			= stmt.fetch!(ubyte)(pos++);
				
				result[i++] = key;
//				delayTask(0);
			}
			
		} catch (Exception e) {
//			count_stmt.finalize();
//			stmt.finalize();
			log(qcount);
			log(query);
			throw e;
		} finally {
			m_db.end();
			release();
		}
		
		return result;
	}
	
	void log(char[] msg)
	{
		Task task = Task.getThis();
		task.errorLog(format("KeyTableSQL: %s", msg));
	}
	
	NyKeyInformation[] searchNameRandom(char[] keyword, uint max = uint.max)
	{
		NyKeyInformation[] result;
		uint i, pos;
		
		if (keyword.length == 0 || nullQuery(keyword)) {
			return null;
		}
		if (keyword[0] == '%' && (keyword.length -1) % 2 == 0) {
			try {
				return searchHash(hexStringToBinary(keyword[1 .. length]), max);
			} catch (ConvError) {
				return null;
			}
			
			return null;
		}
		
		char[] query 		= makeQueryRandom(keyword);
		char[] qcount		= makeCountQueryRandom(keyword);
		signal();
		m_db.begin();
		try {
			uint random_num		= (rand() % 0x10000) + 1;
			
			uint count;
			auto SQLiteStatement count_stmt = m_db.prepare(qcount);
			count_stmt.bind(1, max);
			count_stmt.step();
			count = count_stmt.fetch!(uint)(0);
			
			if (count > max) {
				count = max;
			}
			result = new NyKeyInformation[count];
			
			auto SQLiteStatement stmt = m_db.prepare(query);
			
			stmt.bind(1, random_num, max);
			
			for (i = 0, stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				pos = 0;
				
				key.hash				= stmt.fetch!(ubyte[])(pos++).dup;
				key.fileSize			= stmt.fetch!(uint)(pos++);
				key.fileName			= stmt.fetch!(char[])(pos++).dup;
				key.trip				= stmt.fetch!(char[])(pos++).dup;
				key.timer				= stmt.fetch!(uint)(pos++);
				key.fileNameMBSLength	= stmt.fetch!(uint)(pos++);
				key.keyAddress 			= stmt.fetch!(uint)(pos++);
				key.keyPort				= stmt.fetch!(ushort)(pos++);
				key.bbsAddress			= stmt.fetch!(uint)(pos++);
				key.bbsPort				= stmt.fetch!(ushort)(pos++);
				key.bbsTrip				= stmt.fetch!(ubyte[])(pos++).dup;
				key.referredBlock		= stmt.fetch!(uint)(pos++);
				key.modifiedTime		= stmt.fetch!(uint)(pos++);
				key.ignore				= stmt.fetch!(ubyte)(pos++);
				key.keyVersion			= stmt.fetch!(ubyte)(pos++);
				
				result[i++] = key;
				delayTask(0);
			}
			if (i != count) {
				mbswritef(derr, "test\n");
			}
		} catch (Exception e) {
//			count_stmt.finalize();
//			stmt.finalize();
			log(qcount);
			log(query);
			throw e;
		} finally {
			m_db.end();
			release();
		}
		
		return result;
	}
	
	NyKeyInformation[] searchTripRandom(char[] trip, uint max = uint.max)
	{
		// not implemented
		return searchTrip(trip, max);
	}
	
	NyKeyInformation[] searchTrip(char[] trip, uint max = uint.max)
	{
		NyKeyInformation[] result;
		uint i, pos;
		
		signal();
		m_db.begin();
		try {
			auto SQLiteStatement stmt = m_db.prepare(KEYTABLE_SERACH_TRIP);
			
			stmt.bind(1, trip, max);
			
			for (stmt.step(); stmt.hasRow; stmt.step()) {
				NyKeyInformation key 	= new NyKeyInformation;
				pos = 0;
				
				key.hash				= stmt.fetch!(ubyte[])(pos++).dup;
				key.fileSize			= stmt.fetch!(uint)(pos++);
				key.fileName			= stmt.fetch!(char[])(pos++).dup;
				key.trip				= stmt.fetch!(char[])(pos++).dup;
				key.timer				= stmt.fetch!(uint)(pos++);
				key.fileNameMBSLength	= stmt.fetch!(uint)(pos++);
				key.keyAddress 			= stmt.fetch!(uint)(pos++);
				key.keyPort				= stmt.fetch!(ushort)(pos++);
				key.bbsAddress			= stmt.fetch!(uint)(pos++);
				key.bbsPort				= stmt.fetch!(ushort)(pos++);
				key.bbsTrip				= stmt.fetch!(ubyte[])(pos++).dup;
				key.referredBlock		= stmt.fetch!(uint)(pos++);
				key.modifiedTime		= stmt.fetch!(uint)(pos++);
				key.ignore				= stmt.fetch!(ubyte)(pos++);
				key.keyVersion			= stmt.fetch!(ubyte)(pos++);
				
				result ~= key;
				if (++i >= max) {
					break;
				}
			}
		} catch (Exception e) {
			log("Exception: searchTrip");
			throw e;
		} finally {
//			stmt.finalize();
			m_db.end();
			release();
		}
		
		return result;
	}
	
	uint length()
	{
		uint count;
		
		signal();
		m_db.begin();
		try {
			auto SQLiteStatement stmt = m_db.prepare(KEYTABLE_COUNT);
			stmt.step();
			if (stmt.hasRow) {
				count = stmt.fetch!(int)(0);
			}
		} catch (Exception e) {
			log("Exception: length");
			throw e;
		} finally {
//			stmt.finalize();
			m_db.end();
			release();
		}
		
		return count;
	}
	
	void clear(uint clear_count)
	{
		signal();
		m_db.begin();
		try {
			auto stmt = m_db.prepare(KEYTABLE_CLEAR_COUNT);
			stmt.bind(1, clear_count);
			stmt.step();
		} catch (Exception e) {
			log("Exception: clear 1");
			throw e;
		} finally {
//			stmt.finalize();
			m_db.commit();
			release();
		}
	}
	
	void clear()
	{
		signal();
		m_db.begin();
		try {
			m_db.exec(KEYTABLE_CLEAR);
		} catch (Exception e) {
			log("Exception: clear 2");
			throw e;
		} finally {
			m_db.commit();
			release();
		}
	}
	
	void tick()
	{
		signal();
		m_db.begin();
		try {
			m_db.exec(KEYTABLE_TTL_DOWN);
			m_db.exec(KEYTABLE_TTL_DIE);
		} catch (Exception e) {
			log("Exception: tick");
			throw e;
		} finally {
			m_db.commit();
			release();
		}
	}
	
	void tick30()
	{
		signal();
		m_db.begin();
		try {
			m_db.exec(KEYTABLE_TTL_DOWN30);
			m_db.exec(KEYTABLE_TTL_DIE);
		} catch (Exception e) {
			log("Exception: tick30");
			throw e;
		} finally {
			m_db.commit();
			release();
		}
	}
}


