﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module common.nodetable;

/*
 *  ノード情報テーブル
 * 
 */


private import std.socket;
private import pes.random;
private import pes.net.winny.all;
private import pes.system.all;
private import common.configuration;
import std.stdio;

// ノード要素
class NodeDetails
{
	bool	debug_set;
	ulong	debug_key;
protected:
	
	bool				m_unknown_node;		// よく知らないノード
	uint				m_correlation;		// 相関性（接続優先順位とは別にした）min = 0
	int					m_priority;			// 接続優先順位
	char[]				m_major;			// アプリケーション名
	ushort				m_minor;			// バージョン
	InternetAddress		m_inet_address;		// インターネット上のアドレス
	InternetAddress		m_report_address;	// ノード自己申告アドレス(NAT内の可能性あり)
	char[][3]			m_clustering_key;	// クラスタリングキーワード×3
	NodeType			m_node_type;		// ノードタイプ(RAW, NAT, DDNS, PORT0)
	uint				m_speed;			// 通信速度
	int					m_sort_key;			// ソート用キー
	
	void updateSortKey()
	{
		m_sort_key = ((priority & 0xff) << 24) 
						| ((correlation & 0xff) << 16) 
						| (m_speed & 0xffff);
		
		m_unknown_node = false; // 知った
	}
	
public:
	invariant {
		if (!(debug_key == 0 || debug_key == ((cast(ulong)m_inet_address.addr) << 32| (cast(ulong)m_inet_address.port) & 0xffff))) {
			writef("Assert--- %s:%s\n", debug_key, ((cast(ulong)m_inet_address.addr) << 32| (cast(ulong)m_inet_address.port) & 0xffff));
			assert(0);
		}
	}
	
	this()
	{
		priority = priorityInit;
		m_unknown_node = true;
	}
	
	this(InternetAddress address)
	{
		inetAddress(address.addr, address.port);
		this();
	}
	
	// sorter
	int opCmp(Object rhs)
	{
		NodeDetails r = cast(NodeDetails)rhs;
		
		return r.m_sort_key - m_sort_key;
	}
	
	// getter
	
	bool isUnknownNode()
	{
		return m_unknown_node;
	}
	
	uint correlation()
	{
		return m_correlation;
	}
	
	int priorityInit()
	{
		return 128;
	}
	
	int priority()
	{
		return m_priority;
	}
	
	char[] major()
	{
		return m_major;
	}
	
	ushort minor()
	{
		return m_minor;
	}
	
	InternetAddress inetAddress()
	{
		return m_inet_address;
	}
	
	InternetAddress reportAddress()
	{
		return m_report_address;
	}
	
	char[][] clusteringKey()
	{
		return m_clustering_key.dup;
	}
	
	char[] clusteringKey(int i)
	in {
		assert(0 <= i && i < 3);
	}
	body {
		return m_clustering_key[i];
	}
	
	uint speed()
	{
		return m_speed;
	}
	
	enum NodeType: uint
	{
		RAW,
		NAT,
		DDNS,
		PORT0,
	}
	
	NodeType type()
	{
		return m_node_type;
	}
	
	int sortKey()
	{
		return m_sort_key;
	}
	
	// 連想配列用キー
	// InternetAddressは不要な情報もくっついているのでこれを使う
	// 64bit int型
	ulong aaKey()
	{
		return ((cast(ulong)inetAddress.addr) << 32)
				// 変動するためやめ これだとNAT内のPORT0は一人限定になっちゃう？
//				| (((cast(ulong)reportAddress.addr) & 0xffff) << 16) 
		| ((cast(ulong)inetAddress.port) & 0xffff);
	}
	
	// setter
	
	void correlation(uint correlation)
	{
		if (correlation > 0xff) {
			correlation = 0xff;
		}
		m_correlation = correlation;
		
		updateSortKey();
	}
	
	void priority(int priority)
	{
		if (priority < 0) {
			priority = 0;
		} else if (priority > 0xff) {
			priority = 0xff;
		}
		m_priority = priority;
		
		updateSortKey();
	}
	
	enum FluctuationLevel: int {
		HI			= 16,
		MIDDLE		= 4,
		LOW			= 1,
	}
	
	void priorityDown(FluctuationLevel level )
	{
		priority = priority - cast(int)level;
	}
	
	void priorityUp(FluctuationLevel level)
	{
		priority = priority + cast(int)level;
	}
	
	void major(char[] major)
	{
		m_major = major;
	}
	
	void minor(ushort minor)
	{
		m_minor = minor;
	}
	
	void inetAddress(InternetAddress addr)
	{
		inetAddress(addr.addr, addr.port);
	}
	
	void inetAddress(uint addr, uint port)
	{
		m_inet_address = new InternetAddress(addr, port);
	}
	
	void reportAddress(InternetAddress addr)
	{
		reportAddress(addr.addr, addr.port);
	}
	
	void reportAddress(uint addr, uint port)
	{
		m_report_address = new InternetAddress(addr, port);
	}
	
	void clusteringKey(int index, char[] word)
	in {
		assert(index < m_clustering_key.length);
	}
	body {
		m_clustering_key[index] = word.dup;
	}
	
	void clusteringKey(char[][] keys)
	in {
		assert(keys.length >= m_clustering_key.length);
	}
	body {
		for (int i = 0; i < m_clustering_key.length; ++i) {
			m_clustering_key[i] = keys[i].dup;
		}
	}
	
	void speed(uint speed)
	{
		m_speed = speed;
		
		updateSortKey();
	}
	
	void type(NodeType type)
	{
		m_node_type = type;
	}
	
	// 上位のセット関数
	// できるだけこれを使う
	void know(Configuration config, 
		InternetAddress node_address, NyProtocolHeader node_header,
		NySpeed node_speed, NyConnectionType node_type, NyNodeDetails node_info)
	{
		major			= node_header.major;
		minor			= node_header.minor;
		speed			= node_speed.speed;
		clusteringKey	= node_info.clusteringKey;
		correlation 	= clusterCorrelationPoint(config.clusteringKey, clusteringKey);
		
		inetAddress(node_address.addr, node_info.port);
		reportAddress(node_info.address, node_info.port);
		
		if (node_type.isPort0) {
			type = NodeType.PORT0;
		} else if (compareNode(inetAddress, reportAddress)) {
			type = NodeType.RAW;
		} else {
			type = NodeType.NAT;
		}
	}
	
	
	// 上流検索リンクになりえるノードか？
	bool isUpstreamNode(uint speed)
	{
		// 500KB/SがWinnyの選択肢に標準でないので * 4.3はだめかも
		if (speed * 0.8 > this.speed 
			|| speed * 4.3 < this.speed)
		{
			return false;
		}
		
		return true;
	}
	
	// 下流検索リンクになる得るノードか？
	bool isDownstreamNode(uint speed)
	{
		if (this.speed * 0.8 >= speed
			|| this.speed * 4.3 <= speed)
		{
			return false;
		}
		
		return true;
	}
	
	bool isSelf(Configuration config)
	{
		InternetAddress self = new InternetAddress(config.address, config.fileSharingPort);
		
		if (compareNode(inetAddress, self)) {
			return true;
		}
		if (reportAddress !is null) {
			if (compareNode(reportAddress, self)) {
				return true;
			}
		}
		
		return false;
	}
	
	bool isPort0()
	{
		return type == NodeType.PORT0;
	}
}

// アドレス比較関数
bool compareNode(InternetAddress lhs, InternetAddress rhs)
{
	return (lhs.addr == rhs.addr && lhs.port == rhs.port) ? true:false;
}

// ノード情報テーブル
// キーはNodeDetails.aaKey()を使う
class NodeTable: Locker!(ulong, NodeDetails)
{
public:
	this(SemaphoreResource resource, semaphore id)
	{
		super(resource, id);
	}
	
	NodeDetails[] searchUpstreamNode(NodeDetails node, uint max)
	{
		NodeDetails[] result = searchUpstreamNode_x(node, max * 5);
		
		shuffle!(NodeDetails)(result);
		
		if (result.length > max) {
			result = result[0 .. max-1];
		}
		
		return result;
	}
	
	NodeDetails[] searchUpstreamNode_x(NodeDetails node, uint max)
	{
		NodeDetails[ulong]	result;
		NodeDetails[]		table;
		
		signal();
		try {
			table = values.sort;
			
			// 速度、クラスタ
			foreach (NodeDetails e; table) {
				if (e.isUpstreamNode(node.speed)
					&& clusterCorrelationPoint(node.clusteringKey, e.clusteringKey) > 0)
				{
					result[e.aaKey] = e;
					if (result.length >= max) {
						return result.values;
					}
				}
			}
			
			// 速度
			foreach (NodeDetails e; table) {
				if (e.isUpstreamNode(node.speed)) {
					result[e.aaKey] = e;
					if (result.length >= max) {
						return result.values;
					}
				}
			}
			
			// なんでも
			foreach (NodeDetails e; table) {
				result[e.aaKey] = e;
				if (result.length >= max) {
					return result.values;
				}
			}
		} finally {
			release();
		}
		
		return result.values;
	}
	
	override void set(ulong key, NodeDetails node)
	{
		signal();
		
		try {
			if (exists(key)) {
				NodeDetails en = get(key);
				if (en is node) {
					return; // 同じインスタンス
				}
				en.priority = node.priority;
			} else {
				super.set(key, node);
			}
		} finally {
			release();
		}
	}
	
	
	void downloadSuccess(InternetAddress address, Configuration config)
	{
		signal();
		try {
			NodeDetails node = new NodeDetails(address);
			if (exists(node.aaKey)) {
				node = get(node.aaKey);
				if (node.isUpstreamNode(config.speed)) {
					node.priorityUp(NodeDetails.FluctuationLevel.MIDDLE);
					node.priorityUp(NodeDetails.FluctuationLevel.LOW);
				} else if (node.speed > config.speed) {
					node.priorityUp(NodeDetails.FluctuationLevel.HI);
				} else {
					node.priority = node.priorityInit;
				}
			} else {
				set(node.aaKey, node);
			}
		} finally {
			release();
		}
	}
}
