﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.net.winny.nodeaddress;

// Winny ノードアドレス変換クラス

private import std.socket;
private import std.mmfile;
private import std.stream;
private import std.ctype;
private import pes.rc4;
private import pes.conv;

class NyNodeAddress
{
protected:
	const int 	SUM_VALUE_POS		= 0;
	const int 	CHECK_SUM_MASK		= 0xff;
	const char	NODE_MARK			= '@';
	const char	*DELIMITER			= ":";
	
	static const ubyte MY_RC4_KEY[] = [
			0x6f, 0x70, 0x69, 0x65,
			0x77, 0x66, 0x36, 0x61,
			0x73, 0x63, 0x78, 0x6c, 0x76
	];
	ubyte[13] m_rc4key;
	
public:
	this()
	{
		m_rc4key[] = MY_RC4_KEY[];
	}
	
	
	// n.n.n.n:s -> @xxxx....
	char[] toHashString(InternetAddress address)
	{
		ubyte[] 	node_address = cast(ubyte[])address.toString; // ok
		uint		check_sum;
		ubyte		sum_value;
		char[]		hash_string;
		
		check_sum = 0;
		foreach (ubyte oct; node_address) {
			check_sum += oct;
		}
		sum_value = cast(ubyte)(check_sum & CHECK_SUM_MASK);
		
		m_rc4key[SUM_VALUE_POS] = sum_value;
		hash_string ~= NODE_MARK;
		hash_string ~= binaryToHexString(sum_value ~ pes.rc4.RC4.crypt(m_rc4key, node_address));
		
		return hash_string;
	}
	
	
	// @xxxx.... -> n.n.n.n:s
	InternetAddress toInternetAddress(char[] hashstring)
	{
		ubyte	sum_value;
		uint	check_sum;
		ubyte[]	enc_address;
		char[]	dec_address;
		char[][]addrport;
		
		if (hashstring.length < 20) { // len("@^") + len("0.0.0.0:0") * 2 = 20
			throw new Exception("Specified hash-string is too small");
		}
		if (hashstring[0] != NODE_MARK) {
			throw new Exception("Specified hash-string is not hash-string of NodeAddress");
		}
		
		sum_value 	= (hexToUByte(hashstring[1]) << 4) | hexToUByte(hashstring[2]);
		enc_address	= hexStringToBinary(hashstring[3 .. length]);
		
		m_rc4key[SUM_VALUE_POS] = sum_value;
		
		dec_address = cast(char[])pes.rc4.RC4.crypt(m_rc4key, enc_address);
		
		check_sum = 0;
		foreach (char c; dec_address) { // char type in D is unsigned and 8bit type
			check_sum += cast(uint)c;
		}
		if ((check_sum & CHECK_SUM_MASK) != sum_value) {
			throw new Exception("toInternetAddress: sum check error");
		}
		
		addrport = std.string.split(dec_address, DELIMITER[0 .. 1]);
		foreach (char c; addrport[0]) {
			if (!(std.ctype.isdigit(c) || c == '.')) {
				// DDNS
//				DNSが使えない環境にいるのでコメントアウト
//				InternetHost host = new InternetHost;
//				if (host.getHostByName(addrport[0])) {
//					return new InternetAddress(host.addrList[0], std.conv.toUshort(addrport[1]));
//				}
				throw new HostException(addrport[0]);
			}
		}
		
		return new InternetAddress(addrport[0], std.conv.toUshort(addrport[1]));
	}
	
	static InternetAddress[] load(char[] filepath)
	{
//		auto MmFile 			mmap 		= new MmFile(filepath);
//		MmFileStream			stream		= new MmFileStream(mmap);
		File					stream		= new File(filepath);
		NyNodeAddress			nodeconv	= new NyNodeAddress;
		InternetAddress[]		nodes		= null;
		
		foreach (char[] line; stream) {
			try {
				InternetAddress inetadr = nodeconv.toInternetAddress(line);
				nodes ~= inetadr;
			} catch (Exception) {
				continue;
			}
		}
		stream.close();
		
		return nodes;
	}
	
	
	static void save(char[] filepath, InternetAddress[] nodes,
									FileMode mode = FileMode.OutNew)
	{
		scope(exit) stream.close();
		File 			stream 		= new File(filepath, mode);
		NyNodeAddress	nodeconv	= new NyNodeAddress;
		
		foreach (InternetAddress inetadr; nodes) {
			stream.writeLine(nodeconv.toHashString(inetadr));
		}
	}
}


debug(pes_winny_nodeaddress)
{
	import std.stdio;
	int main()
	{
		InternetAddress adr	= new InternetAddress("123.1.2.3", 1234);
		NyNodeAddress nadr 	= new NyNodeAddress;
		char[] hash 		= nadr.toHashString(adr);
		
		writef("%s\n", hash);
		writef("%s\n", nadr.toInternetAddress(hash));
		
		InternetAddress[] array = NyNodeAddress.load("Noderef2.txt");
		
		// load from file
		foreach (InternetAddress id;  array) {
			writef("%s\n", id);
		}
		
		NyNodeAddress.save("Noderef3.txt", array);
		
		return 0;
	}
}

