﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.locker;

// 
// 無駄なくらいロックするマルチスレッド対応ワンコインロッカー(MAP)
// 
// 
// try-finallyは、
// 代入演算子や比較演算子がオーバーロードされていて
// 例外も考えられるかナーといちおういれておいた
// 
// lock/unlockは
// 長いループ等で明示的にロックしたいときに使う
// ロッカーを開けっ放しでゴソゴソする感じ
// 
// unlockはスレッドIDベースのチェックしかしていないので
// IDが短い期間で再利用されるような環境ではデバッグに注意
// 
// ソートはvaluesで取り出したやつをソートすべし

private import	std.thread;
private import	pes.system.semaphore;
private import 	std.stdio;
private import	pes.hashtable;

debug(locker)
{
	import std.stdio;
}

enum LockerMode
{
	NON,
	AUTO,
}

class Locker(KT, VT)
{
public:
	enum LockType 
	{
		AUTOMATIC,
		MANUAL,
	}
	
protected:
	HashTable!(KT, VT)	m_aa;
	
public:
	this()
	{
		m_aa		= new HashTable!(KT, VT);
	}
	
	Object lockObject(LockerMode mode = LockerMode.AUTO)
	{
		if (mode == LockerMode.AUTO) {
			return this;
		}
		
		return new Object; // dummy
	}
	
	void set(KT key, VT value, LockerMode mode = LockerMode.AUTO)
	{
		synchronized(lockObject(mode)) {
			m_aa.set(key, value);
		}
	}
	
	VT get(KT key, LockerMode mode = LockerMode.AUTO)
	{
		VT value;
		
		synchronized(lockObject(mode)) {
			value = m_aa.get(key);
		}
		
		return value;
	}
	
	// これは取得中だけしかロックしないので操作中に不整合が起こりうるため
	// 普通は、明示的なlock後にLockerMode.NONで取得して、操作後に明示的にunlockするべし
	VT[] values(LockerMode mode = LockerMode.AUTO)
	{
		VT[] values;
		
		synchronized(lockObject(mode)) {
			// aaの実装上valuesはコピーされた配列が戻るため
			// ここではコピーを作らない
			values = m_aa.values;
		}
		
		return values;
	}
	
	KT[] keys(LockerMode mode = LockerMode.AUTO)
	{
		KT[] keys;
		
		synchronized(lockObject(mode)) {
			keys = m_aa.keys;
		}
		
		return keys;
	}
	
	uint length(LockerMode mode = LockerMode.AUTO)
	{
		uint length;
		
		synchronized(lockObject(mode)) {
			length = m_aa.length;
		}
		
		return length;
	}
	
	bool exist(KT key, LockerMode mode = LockerMode.AUTO)
	{
		bool found;
		
		synchronized(lockObject(mode)) {
			found = m_aa.exist(key);
		}
		
		return found;
	}
	
	void remove(KT key, LockerMode mode = LockerMode.AUTO)
	{
		synchronized(lockObject(mode)) {
			m_aa.remove(key);
		}
	}
	
	void clear(LockerMode mode = LockerMode.AUTO)
	{
		synchronized(lockObject(mode)) {
			foreach (KT e; m_aa.keys) {
				remove(e, LockerMode.NON);
			}
		}
	}
	
	bool empty(LockerMode mode = LockerMode.AUTO)
	{
		bool is_empty;
		
		synchronized(lockObject(mode)) {
			is_empty = m_aa.length == 0 ? true: false;
		}
		
		return is_empty;
	}
}
