﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module pes.system.semaphore;
// セマフォ

private import std.thread;
private import pes.system.threadpool;
private import pes.system.delay;
private import pes.system.task;

public alias	uint	semaphore;

public SemaphoreResource	GlobalSemaphore;

static this()
{
	GlobalSemaphore = new SemaphoreResource;
}

private import std.stdio;

private class SemaphoreCounter
{
protected:
	long m_counter;
	uint m_thread_id;
	
public:
	this()
	{
		m_counter = 0;
	}
	
	void signal()
	in {
		assert(!(m_counter != 0 && m_thread_id != Task.getThis.id));
		assert(!(m_counter == 0 && m_thread_id != m_thread_id.init));
	}
	body {
		m_thread_id = Task.getThis.id;
		++m_counter;
	}
	
	void release()
	in {
		assert(!(m_counter != 0 && m_thread_id != Task.getThis.id));
		assert(m_counter > 0);
	}
	body {
		--m_counter;
		if (m_counter < 0) {
			assert(0);
		}
		if (m_counter == 0) {
			m_thread_id = m_thread_id.init;
		}
	}
	
	bool wait()
	{
		if (m_counter == 0) {
			return false;
		}
		
		return true;
	}
	
	bool user()
	{
		return m_thread_id == Task.getThis.id;
	}
	
	long counter()
	{
		return m_counter;
	}
}

class SemaphoreResource
{
protected:
	SemaphoreCounter[semaphore] 	m_semaphore;
	ThreadPool[semaphore]			m_thread_queue;
	
public:
	
	this()
	{
	}
	
	void create(semaphore id)
	{
		synchronized {
			if ((id in m_semaphore) == null) {
				m_semaphore[id] 	= new SemaphoreCounter;
				m_thread_queue[id]	= new ThreadPool;
			}
		}
	}
	
	void signal(semaphore id)
	{
		Task 	self 	= Task.getThis();
		
		synchronized(m_semaphore[id]) {
			if (m_semaphore[id].user) {
				m_semaphore[id].signal();
				return; // ok
			}
			if (m_thread_queue[id].empty) {
				if (! m_semaphore[id].wait) {
					m_semaphore[id].signal();
					return; // ok
				}
			}
			m_thread_queue[id].push(self);
		}
		self.pause();
		m_semaphore[id].signal();
	}
	
	void release(semaphore id)
	{
		Task	task;
		Task	self = Task.getThis;
		
		synchronized(m_semaphore[id]) {
			m_semaphore[id].release();
			
			if ((!m_thread_queue[id].empty) && (! m_semaphore[id].wait)) {
				task 	= cast(Task)m_thread_queue[id].shift;
				if (task is null) {
					throw new TaskException("cast error Task");
				}
				while (! task.tryResume) {
					self.yield();
				}
				while (! m_semaphore[id].wait) {
					self.yield();
				}
			}
		}
	}
	
	semaphore[] ids()
	{
		return m_semaphore.keys;
	}
	
	long counter(semaphore id)
	{
		return m_semaphore[id].counter;
	}
}

debug(pes_system_semaphore)
{
	import	std.stdio;
	import	std.c.time;
	
	enum: semaphore {
		SEM1,
	}
	
	class Data {
		int data;
		this (int i)
		{
			data = i;
		}
		override char[] toString()
		{
			return std.string.toString(data);
		}
	}
	
	void processing()
	{
		uint max = rand() % 10000;
		uint hoge;
		
		for (uint i = 0; i < max; ++i) {
			hoge *= cast(int)((i * 2.34) / 3.12);
		}
	}
	
	class Task1: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				processing();
				if (i == 0) {
					msleep(1000);
				}
				processing();
				writef("task1: get, release %d\n", i);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task2: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				processing();
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				processing();
				signalSemaphore(SEM1);
				processing();
				writef("task2: get, release %d\n", i);
				releaseSemaphore(SEM1);
				processing();
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task3: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				processing();
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				processing();
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				signalSemaphore(SEM1);
				processing();
				writef("task3: get, release %d\n", i);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				processing();
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	class Task4: Task
	{
		this(uint id, TaskResource resource)
		{
			super(id, resource);
		}
		
		int mainLoop()
		{
			
			for (int i = 0; i < 100000; ++i) {
//				delayTask(rand() % 10);
				signalSemaphore(SEM1);
				writef("task4: get, release %d!!\n", i);
				releaseSemaphore(SEM1);
			}
			
			return 0;
		}
		
	}
	
	import std.random;
	
	int main()
	{
		TaskResource	resource = new TaskResource;
		resource.semaphore.create(SEM1);
		
		Task t1 = new Task1(0, resource);
		Task t2 = new Task2(0, resource);
		Task t3 = new Task3(0, resource);
		Task t4 = new Task4(0, resource);
		
		t1.start();
		t2.start();
		t3.start();
		t4.start();
		
		
		t1.wait();
		t2.wait();
		t3.wait();
		
		t4.wait();
		
		
		return 0;
	}
}

