﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.guitask;

// UIタスクの Windows GUI実装

// comple options
// version: POENY 		-> ポエニー
// version: TROPY		-> PoenTropy

private import std.string;
private import std.socket;
private import std.c.time;
private import std.c.stdlib;
private import std.conv;
private import std.windows.charset;
private import std.gc;
private import std.regexp;

private import dfl.all;
private import dfl.winapi;

private import pes.system.all;
private import pes.net.winny.all;
private import pes.net.util;
private import pes.arraybox;
private import pes.conv;
private import pes.variant;
private import common.task;
private import common.environment;
private import task.uitask;
private import task.uicommand;

private import std.stdio;

alias common.environment.Environment Environment;

pragma(lib, "dfl.lib"); // DFL

/**
 * GUI タスク
 *
 */

private static char[] MAIN_FORM_TITLE;	/* メインタイトル	*/

// main.dで分けてほうがいいかも･･･
version(POENY) 
{
	/* ポエニーGUIタスク */
	
	static this()
	{
		MAIN_FORM_TITLE = std.string.format("poeny %04d%02d%02d", 2007, 10, 10);
	}
	
	class GraphicalUserInterfaceTask: UserInterfaceTask
	{
	protected:
		Timer m_timer;
		PoenyMainForm m_main;
		time_t	m_last_gc_time;
		
		override void initialize()
		{
			gcLimit = 4;
			try {
				// スレッド優先順位をINCREASEにする
				//setPriority(PRIORITY.INCREASE);
				
				createResource();
//				loadConfiguration();
				loadDownloadItems();
				m_timer.start();
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			}
			try  {
				if (!startTask()) {
					throw new Error("タスクの起動に失敗:再チャレンジ求む");
				}
				
				environ.downloadList.signal();
				try {
					foreach (DownloadItem e; environ.downloadList.values) {
						addSearch("%" ~ binaryToHexString(e.hash));
					}
				} finally {
					environ.downloadList.release();
				}
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				throw e;
			}
		}
		
		override void terminate()
		{
			try {
				m_timer.stop();
				saveDownloadItems();
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			} finally {
				endTask();
			}
		}
		
		void tick(Object sender, EventArgs ea)
		{
			// タイマーを一時止める
			m_timer.stop();
			m_main.m_task.stopRefreshTimer();
			m_main.m_node.stopRefreshTimer();
			m_main.m_maintenance.stopTimer();
			
			time_t diff = time(null) - m_last_gc_time;
			if (diff > 12) {
	//			log(format("late GC %s", cast(uint)diff));
			}
			
			if (checkEndTask) {
				// 想定外のバグ
				msgBox("誰かタスクがいなくなったようです\n終了します",
													"Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
	//			endTask();
	//			createResource();
	//			startTask();
				exit(0);
			}
			try {
				saveDownloadItems();
			} catch (Object) {
				// do nothing;
			}
			fullCollect();
			
			m_timer.start();
			if (m_main.m_task_timer) {
				m_main.m_task.startRefreshTimer();
			}
			if (m_main.m_node_timer) {
				m_main.m_node.startRefreshTimer();
			}
			if (m_main.m_maintenance_timer) {
				m_main.m_maintenance.startTimer();
			}
			m_last_gc_time = time(null);
		}
		
		override ReturnCode control()
		{
			ReturnCode result = ReturnCode.SUCCESS;
			
			try
			{
				Application.enableVisualStyles();
				m_main = new PoenyMainForm(this);
				Application.run(m_main);
			}
			catch(Object o)
			{
				msgBox(o.toString(), "Fatal Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				
				result = ReturnCode.FAILURE;
			}
			
			return result;
		}
		
	public:
		this(uint uiport, uint ny_port)
		{
			super(uiport, ny_port);
			m_timer = new Timer;
			m_last_gc_time = time(null);
			
			m_timer.tick	~= &tick;
			m_timer.interval = 10_000;
		}
	}
}

version(POEN_TROPY)
{
/* PoenTropy GUIタスク	*/
	
	static this()
	{
		MAIN_FORM_TITLE = std.string.format("PoenTropy α1.4");
	}
	
	class GraphicalUserInterfaceTask: UserInterfaceTask
	{
	protected:
		Timer m_timer;
		PoenTropyMainForm m_main;
		time_t	m_last_gc_time;
		
		override void initialize()
		{
			gcLimit = 4;
			try {
				// スレッド優先順位をINCREASEにする
				//setPriority(PRIORITY.INCREASE);
				
				createResource();
//				loadDownloadItems();
				m_timer.start();
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			}
			try  {
				if (!startTask()) {
					throw new Error("タスクの起動に失敗:再チャレンジ求む");
				}
				
				// 自動検索 追加
				addSearch("【合法】");
				addSearch("ポエム");
				addSearch("過去ログ");
				addSearch("レシピ");
				addSearch("小説");
				addSearch("ジョーク");
				addSearch("日記");
//				addSearch("txt");
				
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				throw e;
			}
		}
		
		override void terminate()
		{
			try {
				m_timer.stop();
//				saveDownloadItems();
			} catch (Exception e) {
				msgBox(e.toString, "Exception", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			} finally {
				endTask();
			}
		}
		
		void tick(Object sender, EventArgs ea)
		{
			// タイマーを一時止める
			m_timer.stop();
			m_main.m_node.stopRefreshTimer();
			m_main.m_maintenance.stopTimer();
			
			
			if (checkEndTask) {
				// 想定外のバグ
				msgBox("誰かタスクがいなくなったようです\n終了します",
													"Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				exit(0);
			}
			// 自動ダウンロード
			
			removeAutoDownload();
			addAutoDownload();
			
			fullCollect();
			
			m_timer.start();
			if (m_main.m_node_timer) {
				m_main.m_node.startRefreshTimer();
			}
			if (m_main.m_maintenance_timer) {
				m_main.m_maintenance.startTimer();
			}
			m_last_gc_time = time(null);
		}
		
		/* 
		 * 自動ダウンローダ
		 * ここに置くべきじゃないけど、べた書きテストの間はここにおいておく
		 */
		
		// 自動ダウンロード追加
		void addAutoDownload()
		{
			bool				found;
			DownloadItem[] 		items;
			DownloadItem		new_item;
			time_t				cur	= time(null);
			NyKeyInformation[]	completed_keys = s_environ.cacheHeader.completeCacheKeys(s_environ.configuration);
			
			// 【合法】and .txt
			s_environ.downloadList.signal();
			s_environ.cacheHeader.signal();
			try {
				NyKeyInformation[] result = s_environ.keyTable.searchName("【合法】");
				foreach (inout NyKeyInformation key; result) {
					if (std.path.getExt(key.fileName) == "txt"
						&& key.fileSize < 0x10000
						&& !s_environ.downloadList.exists(key.hash))
					{
						found = false;
						foreach (NyKeyInformation ckey; completed_keys) {
							if (ckey.hash == key.hash
								&& ckey.fileSize == key.fileSize)
							{
								found = true;
								break;
							}
						}
						if (!found) {
							// "【合法】.*\.txt$" and ファイルサイズ < 64KB and 未登録
							new_item = new DownloadItem(key.hash.dup, key.fileSize, key.fileName.dup);
							new_item.autoDownloadItem = true;		// 自動ダウンロード識別子
							new_item.ttl = cur + key.timer;		 	// リストから消す時間を設定
							
							addDownload(new_item);
							
							items ~= new_item;
						}
					}
					
					key.destroy(); // 破壊
					delete key;
					key = null;
				}
				// delete･･･
				foreach (inout NyKeyInformation key; completed_keys) {
					key.destroy();
					//delete key;
					key = null;
				}
				delete result;
				delete completed_keys;
				result = null;
				completed_keys = null;
			} finally {
				s_environ.downloadList.release();
				s_environ.cacheHeader.release();
			}
			
			// ダウンロードリストへ追加
			foreach (DownloadItem item; items) {
				addDownload(item);
				delayTask(10);
			}
			
		}
		
		// 生存期間切れのキーを自動ダウンロードから取り除く
		void removeAutoDownload()
		{
			DownloadItem[] auto_down;
			time_t			cur = time(null);
			
			s_environ.downloadList.signal();
			s_environ.keyTable.signal();
			try {
				foreach (DownloadItem item; s_environ.downloadList.values) {
					if (item.isAutoDownloadItem)
					{
						// 自動ダウンロードアイテム
						if (item.ttl < cur) {
							// 生存期間切れした
							auto_down ~= item;
						} else if (item.isComplete) {
							// DL完了している
							auto_down ~= item;
						}
					}
				}
			} finally {
				s_environ.downloadList.release();
				s_environ.keyTable.release();
			}
			// 削除 (キーが更新されている場合は、再度登録される)
			foreach (DownloadItem item; auto_down) {
				if (!item.isDownloading) {
					// Download中のアイテムは消さない(終わってから消す)
					removeDownload(item);
					delayTask(10);
				}
			}
		}
		
		override ReturnCode control()
		{
			ReturnCode result = ReturnCode.SUCCESS;
			
			try
			{
				Application.enableVisualStyles();
				m_main = new PoenTropyMainForm(this);
				Application.run(m_main);
			}
			catch(Object o)
			{
				msgBox(o.toString(), "Fatal Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				
				result = ReturnCode.FAILURE;
			}
			
			return result;
		}
		
	public:
		this(uint uiport, uint ny_port)
		{
			super(uiport, ny_port);
			m_timer = new Timer;
			m_last_gc_time = time(null);
			
			m_timer.tick	~= &tick;
			m_timer.interval = 10_000;
		}
	}
}

// 以下クラスなど

void notImplementedDialog()
{
		msgBox("（´・ω・｀）･･未実装のようです", "NotImplemted",
			MsgBoxButtons.OK, MsgBoxIcon.ERROR);
}

char[] comma(uint num)
{
	int i;
	uint n;
	char[] s;
	
	i = 0;
	while(num) {
		n = num % 10;
		num /= 10;
		if (++i == 4) {
			s = "," ~ s, i = 1;
		}
		s = std.string.toString(n) ~ s;
	}
	if (s.length == 0) {
		return "0";
	}
	
	return s;
}

char[] fileSizeString(uint file_size)
{
	double estimate;
	char[] unit;
	
	if (file_size < 1024) {
		estimate	= file_size;
		unit     	= "B";
	} else if (file_size < 1024 * 1000) {
		estimate	= cast(double)file_size / 1024;
		unit		= "KB";
	} else if (file_size < 1024 * 1000 * 1000) {
		estimate	= cast(double)file_size / (1024 * 1000);
		unit		= "MB";
	} else {
		estimate	= cast(double)file_size / (1024 * 1000 * 1000);
		unit		= "GB";
	}
	
	return std.string.format("%.1f%s", estimate, unit);
}

// 数値入力用テキストボックス
class NumericTextBox: TextBox
{
	override void wndProc(inout Message msg)
	{
		switch (msg.msg) {
		case WM_CHAR:
			if (!('0' <= msg.wParam && msg.wParam <= '9')
				&& msg.wParam != '\x08' 
				&& msg.wParam != '\x7f')
			{
				msg.result = false;
				return;
			}
		default:
			super.wndProc(msg);
		}
	}
}

class SortListView: ListView
{
	abstract UIFileSortType sortType();
	abstract bool asc();
}


// 検索タブ
class SearchTabPage: TabPage
{
protected:
	// ENTERを受け付けるテキストボックス
	class CommandTextBox: TextBox
	{
	protected:
		alias void delegate() EnterCommandCallBack;
		EnterCommandCallBack m_enter_callback;
		override void wndProc(inout Message msg)
		{
			switch (msg.msg) {
			case WM_KEYDOWN:
				if (msg.wParam == Keys.ENTER) {
					m_enter_callback();
					msg.result = true;
					break;
				}
				super.wndProc(msg);
				break;
			case WM_GETDLGCODE:
				msg.result |= DLGC_WANTALLKEYS;
				break;
			default:
				super.wndProc(msg);
			}
		}
	public:
		this(EnterCommandCallBack enterevent)
		{
			m_enter_callback 	= enterevent;
		}
	}
	
	class ListViewKeyItem {
		UIFileInformation[] keys;
		ListView			view;
		this(ListView v, UIFileInformation[] info)
		{
			view = v;
			keys = info;
		}
		
		void clear()
		{
			if (keys.length == 0) {
				return;
			}
			foreach (inout UIFileInformation e; keys) {
				e.destroy();
				delete e;
				e = null;
			}
			delete keys;
			keys = null;
			s_parent.fullCollect();
		}
	}
	
	class SearchResultView: SortListView
	{
		MenuItem m_download;
		MenuItem m_details;
		
		void menuPopup(Object sender, EventArgs ea)
		{
			if (selectedItems.count == 0) {
				m_download.enabled	= false;
				m_details.enabled	= false;
			} else {
				m_download.enabled	= true;
//				m_details.enabled	= true;
				m_details.enabled	= false;// 未実装
			}
		}
		
		void itemDownload(Object sender, EventArgs ea)
		{
			UIFileInformation 	info;
			DownloadItem		item;
			char[] name;
			TabPage page = m_search.selectedTab;
			
			if (page is null) {
				return;
			}
			
			if (selectedIndices.count == 0) {
				return;
			}
			
			foreach (int e; selectedIndices) {
				info = m_pages[page.text].keys[e];
				item = new DownloadItem(info.hash.dup, info.fileSize, info.fileName.dup);
				addDownload(item);
			}
		}
		
		void itemDetails(Object sender, EventArgs ea)
		{
			notImplementedDialog();
		}
		
		void initialize()
		{
			cmenu = new ContextMenu;
			cmenu.popup ~= &menuPopup;
			
			with (m_download = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("ダウンロード");
				click ~= &itemDownload;
				index = 0;
			}
			cmenu.menuItems.add(m_download);
			
			with (m_details = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("詳細表示");
				click ~= &itemDetails;
				index = 1;
			}
			cmenu.menuItems.add(m_details);
		}
		
		this()
		{
			initialize();
		}
		
		// リストのソート
		UIFileSortType 	m_type = UIFileSortType.FILE_SIZE;
		bool			m_asc = false;
		
		override UIFileSortType sortType()
		{
			return m_type;
		}
		
		override bool asc()
		{
			return m_asc;
		}
		
		protected override void onReflectedMessage(inout Message m)
		{
			struct NM_LISTVIEW {
				NMHDR hdr;
				int iItem;
				int iSubItem;
			}
			super.onReflectedMessage(m);
			
			const uint LVN_COLUMNCLICK =         (LVN_FIRST-8);
			NMHDR* nmh;
			nmh = cast(NMHDR*)m.lParam;
			if (m.msg != WM_NOTIFY || nmh.code != LVN_COLUMNCLICK) {
				return;
			}
			
			NM_LISTVIEW *col = cast(NM_LISTVIEW *)m.lParam;
			
			clickColumn(col.iSubItem);
		}
		
		void clickColumn(int index)
		{
			UIFileSortType newtype;
			
			switch (index) {
			case 0: // ファイル名
				newtype = UIFileSortType.FILE_NAME;
				break;
			case 1: // 拡張子
				newtype = UIFileSortType.FILE_EXT;
				break;
			case 2: // サイズ
				newtype = UIFileSortType.FILE_SIZE;
				break;
			case 3: // 人気度
				newtype = UIFileSortType.REFF_COUNT;
				break;
			}
			if (m_type != newtype) {
				m_type = newtype;
				m_asc = m_asc;
			} else {
				m_type = newtype;
				m_asc = !m_asc;
			}
			
			refreshSearch(m_search.selectedTab, false, m_type, m_asc);
		}
	}
	
	UserInterfaceTask		s_parent;
	
	Environment 			s_environ;
	TabControl 				m_search;
	ListViewKeyItem[char[]]	m_pages;
	Button					m_button;
	TextBox					m_input;
	Label					m_results;
	time_t					m_last_search_time;
	Panel					m_panel;
	
	ListView viewCreate()
	{
		return new SearchResultView;
	}
	
	// 画面初期化
	void initialize()
	{
		with (m_input = new CommandTextBox(&addSearch))
		{
			dock 			= DockStyle.NONE;
			bounds			= Rect(10, 5, 250, 20);
			parent			= this;
		}
		
		with (m_button = new Button)
		{
			dock 	= DockStyle.NONE;
			text	= "検索";
			bounds	= Rect(260, 5, 80, 20);
			parent	= this;
			click	~= &buttonClick;
		}
		
		with (m_results = new Label) {
			dock 			= DockStyle.NONE;
			text 	= "hit(0)";
			bounds	= Rect(345, 9, 80, 20);
//			bounds	= Rect(135, 9, 80, 20);
			parent = this;
		}
		
		// control
		with (m_search = new TabControl)
		{
			dock 	= DockStyle.NONE;
//			size	= Size(100, 100);
			bounds	= Rect(0, 30, 490, 400);
			parent 	= this;
			
			mouseDown 				~= &mouseClick;
			selectedIndexChanged 	~= &this.selectedIndexChanged;
		}
		resize ~= &resizeSearchTab;
	}
	
	void resizeSearchTab(Object sender, EventArgs ea)
	{
		Size current_size = size;
		m_search.bounds = Rect(0, 30, current_size.width, current_size.height - 30);
	}
	
	void selectedIndexChanged(Object sender, EventArgs ea)
	{
		refreshSearch(m_search.selectedTab);
	}
	
	char[][] columnFile(UIFileInformation info)
	{
		char[][] row = new char[][4];
		
		row[0] = info.fileName;
		row[1] = std.path.getExt(info.fileName);
		row[2] = fileSizeString(info.fileSize);
		row[3] = std.string.format("%s", 
			(info.referredCount * 0x10000) / ((info.fileSize < 0x10000) ? 0x10000:info.fileSize));
		
		return row;
	}
	
	// 検索画面リフレッシュ
	void refreshSearch(TabPage page,
			bool search_real = true, UIFileSortType type = UIFileSortType.FILE_SIZE,
			bool asc = false)
	{
		UIRSearch 			search;
		UIFileInformation[] rows;
		ListView			listview;
		int					index;
		
		if (page is null) {
			return;
		}
		
		m_results.text = std.string.format("hit(%d)", m_pages[page.text].view.items.count);
		
		// 頻繁な更新は避ける
		if (m_last_search_time + 1 > time(null)) {
			return;
		}
		m_last_search_time = time(null);
		
		listview = m_pages[page.text].view;
		if (search_real) {
			search = searchFile(page.text, type);
			rows = search.record;
		} else {
			rows = m_pages[page.text].keys;
		}
		UIFileInformation.setSortType(rows, type);
		UIFileInformation.setSortAscent(rows, asc);
		rows = rows.sort;
		
		listview.beginUpdate();
		listview.clear();
		
		index = 0;
		foreach (UIFileInformation e; rows) {
			listview.addRow(columnFile(e));
		}
		m_results.text = std.string.format("hit(%d)", rows.length);
		
		listview.endUpdate();
		listview.update();
		if (search_real) {
			m_pages[page.text].clear();
		}
		m_pages[page.text].keys = rows;
	}
	
	// 検索ボタンクリック
	void onEnterInput(Object sender, KeyEventArgs ea)
	{
		if (ea.keyCode == Keys.ENTER) {
			addSearch();
		}
	}
	
	// 検索ボタンクリック
	void buttonClick(Object sender, EventArgs ea)
	{
		addSearch();
	}
	
	void addSearch(char[] key)
	{
		s_parent.addSearch(key);
	}
	
	void removeSearch(char[] key)
	{
		s_parent.removeSearch(key);
	}
	
	void addDownload(DownloadItem item)
	{
		s_parent.addDownload(item);
	}
	
	void itemDoubleClick(Object sender, EventArgs ea)
	{
		ListView view = (cast(ListView)sender);
		UIFileInformation item;
		char[] name;
		uint fileSize, referred;
		TabPage page = m_search.selectedTab;
		
		if (page is null || view is null) {
			return;
		}
		
		if (view.selectedIndices.count == 0) {
			return;
		}
		
		item = m_pages[page.text].keys[view.selectedIndices[0]];
		
		addDownload(new DownloadItem(item.hash.dup, item.fileSize, item.fileName.dup));
		// ダウンロード登録
	}
	
	// 検索登録
	void addSearch()
	{
		ListView listview;
		if (m_input.text.length == 0) {
			return;
		}
		char[] key = m_input.text.dup;
		
		key = std.string.replace(key, "　", " ");
		
		if ((key in m_pages)) {
			return;
		}
		
		TabPage page = new TabPage(key);
		page.parent = m_search;
		with (listview = viewCreate) {
			ColumnHeader column;
			
			dock 			= DockStyle.FILL;
			font 			= new Font("MS ゴシック", 9f);
			foreColor 		= Color(10, 10, 20);
			backColor		= Color(255, 250, 250);
			view 			= View.DETAILS;
			fullRowSelect 	= true;
			gridLines 		= true;
			doubleClick		~= &itemDoubleClick;
//			sorting			= SortOrder.DESCENDING;
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("ファイル"); //日本語最悪です
				width = 305;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text	= cast(char[])utf8ToMBS("タイプ");
				textAlign	= HorizontalAlignment.CENTER;
				width = 45;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text	= cast(char[])utf8ToMBS("サイズ");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 55;
			}
			columns.add(column);
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("人気度");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 55;
			}
			columns.add(column);
			parent = page;
		}
		
		m_pages[key] = new ListViewKeyItem(listview, null);
		m_input.text = "";	// 入力を消す
		
		m_search.tabPages.add(page);
		m_search.selectedTab = page;	// 選択
		refreshSearch(page);			// 更新
		
		// クエリタスクへ登録
		addSearch(key);
	}
	
	// 検索
	UIRSearch searchFile(char[] key, UIFileSortType sort_type)
	{
		Mail				mail;
		UIRSearch	 		result;
		UIQSearch			search 	= new UIQSearch;
		
		search.key = key;
		search.keyType =  UISearchKeyType.KEYWORD;
		
		result	= searchFileUI(search);
		
		result.max = search.max;
		result.sort(sort_type);
		
		search = null;
		
		return result;
	}
	
	// UI用検索 
	UIRSearch searchFileUI(UIQSearch search)
	{
		UIRSearch 			result = new UIRSearch;
		UIFileInformation[]	array;
		UIFileInformation	info;
		NyKeyInformation	keys[];
		uint i;
		
		switch (search.keyType) {
		case UISearchKeyType.KEYWORD:
			keys = s_environ.keyTable.searchName(search.key);
			break;
		case UISearchKeyType.TRIP:
			keys = s_environ.keyTable.searchTrip(search.key);
			break;
		case UISearchKeyType.HASH:
			ubyte[] hash = hexStringToBinary(search.key);
			keys = s_environ.keyTable.searchHash(hash);
			break;
		}
		if (keys.length == 0) {
			return result;
		}
		
		ubyte[][] itemkey = s_environ.downloadList.keys();
		uint n = 0; bool found = false;
		foreach (NyKeyInformation e; keys) {
			if ((!(e.keyAddress == s_environ.configuration.address
					&& e.keyPort == s_environ.configuration.fileSharingPort)))
			{
				found = false;
				foreach (ubyte[] hash; itemkey) {
					if (hash == e.hash) {
						found = true;
						break;
					}
				}
				if (found == false) {
					++n;
				}
			}
		}
		array = new UIFileInformation[n];
		n = 0; 
		foreach (NyKeyInformation e; keys) {
			if ((!(e.keyAddress == s_environ.configuration.address
					&& e.keyPort == s_environ.configuration.fileSharingPort)))
			{
				found = false;
				foreach (ubyte[] hash; itemkey) {
					if (hash == e.hash) {
						found = true;
						break;
					}
				}
				if (found == false) {
					info = new UIFileInformation;
					
					info.fileName 		= e.fileName.dup;
					info.fileSize		= e.fileSize;
					info.hash			= e.hash;
					info.trip			= e.trip;
					info.referredCount	= e.referredBlock;
					
					array[n++] = info;
				}
			}
		}
		
		result.record(array);
		foreach (inout NyKeyInformation e; keys) {
			e.destroy();
			delete e;
			e = null;
		}
		delete keys;
		keys = null;
		array = null;
		info = null;
		
		return result;
	}
	
	void mouseClick(Object sender, MouseEventArgs ea)
	{
		TabPage page;
		if(ea.button == MouseButtons.RIGHT) {
			page = m_search.selectedTab;
			if (page !is null) {
				// 検索ワード削除
				m_pages[page.text].clear();
				m_pages.remove(page.text);
				m_search.tabPages.remove(page);
				
				// クエリタスクからも削除
				removeSearch(page.text);
			}
		} else if (ea.button == MouseButtons.LEFT) {
			page = m_search.selectedTab;
			if (page !is null) {
				SortListView view = cast(SortListView)m_pages[page.text].view;
				// ページ更新
				refreshSearch(page, true, view.sortType, view.asc);
			}
		}
	}
	
public:
	this(UserInterfaceTask parent)
	{
		super("検索");
		initialize();
		s_environ 	= parent.environ;
		s_parent	= parent;
	}
	
	void setFocus()
	{
		m_input.focus();
	}
}


// ダウンロードタブ
class DownloadTabPage: TabPage
{
protected:
	DownloadItem[]	m_items;
	
	class DownloadItemView: ListView
	{
		MenuItem m_remove;
		MenuItem m_decode;
		
		void menuPopup(Object sender, EventArgs ea)
		{
			bool downloading = false;
			
			if (selectedItems.count == 0) {
				m_remove.enabled = false;
				m_decode.enabled = false;
			} else {
				m_remove.enabled = true;
				m_decode.enabled = true;
				
//				foreach (int e; selectedIndices) {
//					if (m_items[s].downloading) {
//						downloading = true;
//						break;
//					}
//				}
//				if (downloading) {
//					m_decode.enabled = false; // DL中のアイテムがある場合は変換できない
//				} else {
//					m_decode.enabled = true;
//				}
			}
		}
		
		void itemRemove(Object sender, EventArgs ea)
		{
			foreach (int e; selectedIndices) {
				removeDownload(m_items[e]);
			}
			
//			refreshItem(); // コントロール側で削除するため更新できない
		}
		
		void itemDecode(Object sender, EventArgs ea)
		{
			foreach (int e; selectedIndices) {
				decodeCache(m_items[e]);
			}
		}
		
		void initialize()
		{
			cmenu = new ContextMenu;
			cmenu.popup ~= &menuPopup;
			
			with (m_decode = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("キャッシュ強制変換");
				click ~= &itemDecode;
				index = 0;
			}
			cmenu.menuItems.add(m_decode);
			
			with (m_remove = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("ダウンロード削除");
				click ~= &itemRemove;
				index = 1;
			}
			cmenu.menuItems.add(m_remove);
		}
		
		this()
		{
			initialize();
		}
	}
	UserInterfaceTask	s_parent;
	DownloadList		s_download_list;
	
	DownloadItemView	m_download_view;
	Button				m_new_item_button;
	Button				m_update_button;
	
	void initialize()
	{
		with(m_download_view = new DownloadItemView)
		{
			ColumnHeader column;
			
			dock 			= DockStyle.NONE;
			bounds			= Rect(0, 50, 490, 375);
			font 			= new Font("MS ゴシック", 9f);
			foreColor 		= Color(10, 10, 20);
			backColor		= Color(255, 250, 250);
			view 			= View.DETAILS;
			fullRowSelect 	= true;
			gridLines 		= true;
//			sorting			= SortOrder.DESCENDING;
			
			with(column = new ColumnHeader)
			{
				column.text	= cast(char[])utf8ToMBS("状況");
				textAlign	= HorizontalAlignment.CENTER;
				width = 50;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("ファイル");
				textAlign	= HorizontalAlignment.LEFT;
				width = 240;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("サイズ");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 80;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("進捗");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 70;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("候補");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 40;
			}
			columns.add(column);
			parent = this;
		}
		resize 			~= &resizeDownloadTab;
		
		with (m_new_item_button = new Button)
		{
			text			= "新規";
			bounds			= Rect(10, 5, 120, 20);
			click			~= &clickNewButton;
			parent			= this;
			enabled			= false;
		}
		
		with (m_update_button = new Button)
		{
			text			= "更新";
			bounds			= Rect(135, 5, 120, 20);
			click 			~= &clickUpdateButton;
			parent			= this;
		}
	}
	
	void removeDownload(DownloadItem item)
	{
		s_parent.removeDownload(item);
	}
	
	void decodeCache(DownloadItem item)
	{
		s_parent.decodeCache(item);
	}
	
	void resizeDownloadTab(Object sender, EventArgs ea)
	{
		Size current_size = size;
		m_download_view.bounds = Rect(0, 50, current_size.width, current_size.height - 50);
	}
	
	void clickNewButton(Object sender, EventArgs ea)
	{
		notImplementedDialog();
	}
	
	void clickUpdateButton(Object sender, EventArgs ea)
	{
		refreshItem();
	}
	
	char[][] columnItem(DownloadItem item)
	{
		char[][] row = new char[][5];
		if (item.isDownloading) {
			row[0] = "DL中";
		} else if (item.isComplete) {
			row[0] = "完了";
		} else {
			row[0]	= "待機";
		}
		row[1]	= item.name.dup;
		row[2]	= std.string.format("%s", comma(item.dataSize));
		row[3]	= std.string.format("%s/%s", item.blockCount, item.blocks);
		row[4]	= std.string.format("%s", item.distributorCount);
		
		return row;
	}
	
public:
	this(UserInterfaceTask parent)
	{
		super("ダウンロード");
		s_parent 		= parent;
		s_download_list	= parent.environ.downloadList;
		initialize();
	}
	
	void refreshItem()
	{
		m_download_view.beginUpdate();
		m_download_view.clear();
		m_items.length = 0;
		
//		s_download_list.signal();
		try {
			foreach (DownloadItem e; s_download_list.values.sort) {
				m_download_view.addRow(columnItem(e));
				m_items ~= e;
			}
		} finally {
//			s_download_list.release();
		}
		
		m_download_view.endUpdate();
	}
}

// キャッシュタブ
class CacheTabPage: SearchTabPage
{
	class BlockRate {
		uint blocks;
		uint count;
		bool completed;
		
		this(uint blocks, uint count, bool completed)
		{
			this.blocks = blocks;
			this.count = count;
			this.completed = completed;
		}
		
		override int opCmp(Object rhs)
		{
			BlockRate rhs_rate = cast(BlockRate)rhs;
			
			return (cast(long)((cast(double)count/blocks)*100)) 
			- (cast(long)((cast(double)rhs_rate.count/rhs_rate.blocks)*100));
		}
	}
	
	class SearchCacheResultView: SortListView
	{
		MenuItem m_download;
		MenuItem m_decode;
		MenuItem m_remove;
		MenuItem m_details;
		
		bool checkDownload()
		{
			UIFileInformation 	info;
			DownloadItem		item;
			char[] name;
			TabPage page = m_search.selectedTab;
			CacheHeader header;
			
			if (page is null) {
				return false;
			}
			
			if (selectedIndices.count == 0) {
				return false;
			}
			
			s_environ.cacheHeader.signal();
			try {
				foreach (int e; selectedIndices) {
					info 	= m_pages[page.text].keys[e];
					if (s_environ.cacheHeader.exists(info.hash, info.fileSize)) {
						header 	= s_environ.cacheHeader.get(info.hash, info.fileSize);
						if (header.completed) {
							return false;
						}
					}
				}
			} finally {
				s_environ.cacheHeader.release();
			}
			
			return true;
		}
		
		void menuPopup(Object sender, EventArgs ea)
		{
			if (selectedItems.count == 0) {
				m_decode.enabled	= false;
				m_remove.enabled	= false;
				m_details.enabled	= false;
				m_download.enabled 	= false;
			} else {
				if (checkDownload()) {
					m_download.enabled = true;
				} else {
					m_download.enabled = false;
				}
				
				m_decode.enabled	= true;
				m_remove.enabled	= true;
//				m_details.enabled	= true;
				m_details.enabled	= false; // 未実装
			}
		}
		
		void itemDownload(Object sender, EventArgs ea)
		{
			UIFileInformation 	info;
			DownloadItem		item;
			char[] name;
			TabPage page = m_search.selectedTab;
			
			if (page is null) {
				return;
			}
			
			if (selectedIndices.count == 0) {
				return;
			}
			
			foreach (int e; selectedIndices) {
				info = m_pages[page.text].keys[e];
				item = new DownloadItem(info.hash.dup, info.fileSize, info.fileName.dup);
				addDownload(item);
			}
		}
		
		void itemDecode(Object sender, EventArgs ea)
		{
			UIFileInformation 	info;
			DownloadItem		item;
			char[] name;
			TabPage page = m_search.selectedTab;
			
			if (page is null) {
				return;
			}
			
			if (selectedIndices.count == 0) {
				return;
			}
			
			foreach (int e; selectedIndices) {
				info = m_pages[page.text].keys[e];
				item = new DownloadItem(info.hash.dup, info.fileSize, info.fileName.dup);
				decodeCache(item);
			}
		}
		
		void itemRemove(Object sender, EventArgs ea)
		{
			UIFileInformation 	info;
			DownloadItem		item;
			char[] name;
			TabPage page = m_search.selectedTab;
			
			if (page is null) {
				return;
			}
			
			if (selectedIndices.count == 0) {
				return;
			}
			
			foreach (int e; selectedIndices) {
				info = m_pages[page.text].keys[e];
				item = new DownloadItem(info.hash.dup, info.fileSize, info.fileName.dup);
				removeCache(item);
			}
		}
		
		void itemDetails(Object sender, EventArgs ea)
		{
			notImplementedDialog();
		}
		
		void initialize()
		{
			cmenu = new ContextMenu;
			cmenu.popup ~= &menuPopup;
			
			with (m_download = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("ダウンロード登録");
				click ~= &itemDownload;
				index = 0;
			}
			cmenu.menuItems.add(m_download);
			
			with (m_decode = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("キャッシュ強制変換");
				click ~= &itemDecode;
				index = 1;
			}
			cmenu.menuItems.add(m_decode);
			
			with (m_remove = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("キャッシュ削除");
				click ~= &itemRemove;
				index = 2;
			}
			cmenu.menuItems.add(m_remove);
			
			with (m_details = new MenuItem)
			{
				text = cast(char[])utf8ToMBS("詳細表示");
				click ~= &itemDetails;
				index = 3;
			}
			cmenu.menuItems.add(m_details);
		}
		
		this()
		{
			initialize();
		}
		
		// リストのソート
		UIFileSortType 	m_type = UIFileSortType.FILE_SIZE;
		bool			m_asc = false;
		
		override UIFileSortType sortType()
		{
			return m_type;
		}
		
		override bool asc()
		{
			return m_asc;
		}
		
		protected override void onReflectedMessage(inout Message m)
		{
			struct NM_LISTVIEW {
				NMHDR hdr;
				int iItem;
				int iSubItem;
			}
			super.onReflectedMessage(m);
			
			const uint LVN_COLUMNCLICK =         (LVN_FIRST-8);
			NMHDR* nmh;
			nmh = cast(NMHDR*)m.lParam;
			if (m.msg != WM_NOTIFY || nmh.code != LVN_COLUMNCLICK) {
				return;
			}
			
			NM_LISTVIEW *col = cast(NM_LISTVIEW *)m.lParam;
			
			clickColumn(col.iSubItem);
		}
		
		void clickColumn(int index)
		{
			UIFileSortType newtype;
			
			switch (index) {
			case 0: // ファイル名
				newtype = UIFileSortType.FILE_NAME;
				break;
			case 1: // 拡張子
				newtype = UIFileSortType.FILE_EXT;
				break;
			case 2: // サイズ
				newtype = UIFileSortType.FILE_SIZE;
				break;
			case 3: // 人気度
				newtype = UIFileSortType.REFF_COUNT;
				break;
			case 4: // 率
				newtype = UIFileSortType.TAG;
				break;
			}
			if (m_type != newtype) {
				m_type = newtype;
				m_asc = m_asc;
			} else {
				m_type = newtype;
				m_asc = !m_asc;
			}
			
			refreshSearch(m_search.selectedTab, false, m_type, m_asc);
		}
	}
	
	void decodeCache(DownloadItem item)
	{
		s_parent.decodeCache(item);
	}
	
	void removeCache(DownloadItem item)
	{
		s_parent.removeCache(item);
	}
	
	this(UserInterfaceTask parent)
	{
		super(parent);
		text = "キャッシュ";
	}
	
	override ListView viewCreate()
	{
		return new SearchCacheResultView;
	}
	
	override void addSearch(char[] key)
	{
		// do nothing
	}
	
	override void removeSearch(char[] key)
	{
		// do nothing
	}
	
	override void itemDoubleClick(Object sender, EventArgs ea)
	{
		super.itemDoubleClick(sender, ea);
	}
	
	override void addSearch()
	{
		char[] 			key	 = m_input.text;
		ListView 		view;
		ColumnHeader	column;
		
		super.addSearch();
		if (!(key in m_pages)) {
			return;
		}
		view = m_pages[key].view;
		
		if (view.columns.count < 5) { // [率]が無い場合は追加
			view.columns[0].width = 260;
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("率");
				textAlign	= HorizontalAlignment.RIGHT;
				width = 40;
			}
			view.columns.add(column);
		}
	}
	
	override char[][] columnFile(UIFileInformation info)
	{
		char[][] row = new char[][5];
		BlockRate rate = cast(BlockRate)info.tag;
		
		row[0] = info.fileName;
		row[1] = std.path.getExt(info.fileName);
		row[2] = std.string.format("%s", fileSizeString(info.fileSize));
		row[3] = std.string.format("%s", 
			(info.referredCount * 0x10000) / ((info.fileSize < 0x10000) ? 0x10000:info.fileSize));
		
		row[4] = std.string.format("%s%%", cast(long)((cast(double)rate.count / rate.blocks) * 100));
		
		return row;
	}
	
	// UI用検索 
	override UIRSearch searchFileUI(UIQSearch search)
	{
		UIRSearch 			result = new UIRSearch;
		UIFileInformation[]	array;
		UIFileInformation	info;
		CacheHeader			keys[];
		ulong				blocks, count;
		uint				i;
		
		switch (search.keyType) {
		case UISearchKeyType.KEYWORD:
			keys = s_environ.cacheHeader.searchName(search.key);
			break;
		}
		array = new UIFileInformation[keys.length];
		i = 0;
		foreach (inout CacheHeader e; keys) {
			blocks 	= (e.dataSize / 0x10000) + (e.dataSize % 0x10000 ? 1:0);
			count	= e.bitmap.count(blocks);
			
			info = new UIFileInformation;
			
			info.fileName 		= e.fileName;
			info.fileSize		= e.dataSize;
			info.hash			= e.hash;
			info.trip			= s_environ.configuration.trip;
			info.referredCount	= e.referredCount;
			info.tag			= new BlockRate(blocks, count, e.completed != 0);
			
			array[i++] = info;
			e = null;
		}
		delete keys;
		keys = null;
		result.record(array);
		
		return result;
	}
}


class PoemTabPage: TabPage
{
protected:
	Environment	s_environ;
	TextBox		m_title_text;
	TextBox		m_poem;
	Button		m_update;
	Button		m_random;
	Button		m_reset;
	
	void initialize()
	{
		Size current_size = size;
		
		with (m_title_text = new TextBox)
		{
			bounds			= Rect(15, 30, 455, 20);
			parent			= this;
		}
		
		with (m_poem = new TextBox)
		{
			bounds			= Rect(15, 60, 455, 320);
			multiline		= true;
			acceptsReturn	= true;
			scrollBars 		= ScrollBars.BOTH;
			parent			= this;
		}
		
		with (m_update = new Button)
		{
			text	= "ポエむ";
			bounds	= Rect(360, 390, 100, 20);
			click	~= &publish;
			parent	= this;
		}
		
		with (m_random = new Button)
		{
			text	= "ランダむ";
			bounds	= Rect(260, 390, 100, 20);
			click	~= &getRandom;
			parent	= this;
		}
		
		with (m_reset = new Button)
		{
			text	= "リセッと";
			bounds	= Rect(160, 390, 100, 20);
			click	~= &resetPoem;
			parent	= this;
		}
		
		resize ~= &resizePoem;
	}
	
	void resizePoem(Object sender, EventArgs ea)
	{
		// 500,480
		Size current_size = size;
		m_title_text.bounds	= Rect(15, 30, current_size.width - 35, 20);
		m_poem.bounds 		= Rect(15, 60, current_size.width - 35, current_size.height - 100);
		m_update.bounds 	= Rect(current_size.width - 125, current_size.height - 32, 100, 20);
		m_random.bounds		= Rect(current_size.width - 225, current_size.height - 32, 100, 20);
		m_reset.bounds		= Rect(current_size.width - 325, current_size.height - 32, 100, 20);
	}
	
	void publish(Object sender, EventArgs ea)
	{
		DialogResult result;
		ubyte[] poem;
		char[] filename;
		
		result = msgBox("このポエムは一度公開する二度と地上から消すことはできません", 
							"WARNING", MsgBoxButtons.OK_CANCEL, MsgBoxIcon.WARNING);
		
		if (result != DialogResult.OK) {
			return;
		}
		
		if (m_title_text.text.length == 0) {
			msgBox("タイトルを考えてください", "Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			return;
		}
		filename = s_environ.configuration.uploadFolder ~ std.file.sep ~ m_title_text.text ~ ".txt";
		
		poem = utf8ToMBS(m_poem.text);
		if (poem.length == 0) {
			msgBox("ポエムが足りません", "Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			return;
		}
		
		try {
			File file = new File(filename, FileMode.OutNew);
			file.writeBlock(poem, poem.length);
		} catch (Exception e) {
			msgBox("エラーでました\n" ~ e.toString, "Error", MsgBoxButtons.OK, MsgBoxIcon.ERROR);
			return;
		}
		
		
		refresh();
	}
	
	void getRandom(Object sender, EventArgs ea)
	{
		char[][] poem_list;
		uint 	index;
		char[]	poem_name;
		char[]	poem_title;
		char[]	poem;
		
		poem_list 	= std.file.listdir(s_environ.configuration.downloadFolder, RegExp(r"【合法】.*\.txt$"));
		if (poem_list.length != 0) {
			index		= rand() % poem_list.length;
			poem_name	= poem_list[index];
			poem_title	= std.path.getName(std.path.getBaseName(poem_name));
			poem 		= mbsToUTF8(cast(ubyte[])std.file.read(poem_name));
			
			m_title_text.text 	= poem_title;
			m_poem.text			= poem;
		}
	}
	
	void resetPoem(Object sender, EventArgs ea)
	{
		refresh();
	}
	
	void refresh()
	{
		char[] old = m_poem.text;
		m_title_text.text = "【合法】【ポエム】 ";
		m_poem.text = "";
		m_title_text.focus();
		
		old = null;
		delete old;
	}
	
	this(Environment environ)
	{
		s_environ = environ;
		
		super("ポエム");
		initialize();
		refresh();
	}
}


class TaskTabPage: TabPage
{
protected:
	UserInterfaceTask	s_parent;
	Environment			s_environ;
	ListView			m_view;
	Timer				m_timer;
	
	void initialize()
	{
		with (m_view = new ListView) {
			ColumnHeader column;
			
			dock 			= DockStyle.FILL;
			font 			= new Font("MS ゴシック", 9f);
			foreColor 		= Color(0);
//			sorting 		= SortOrder.ASCENDING;
			view 			= View.DETAILS;
			fullRowSelect 	= true;
			gridLines 		= true;
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("タイプ");
				width = 80;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("レート(KB/S)");
				width = 85;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("率(%)");
				width = 50;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = cast(char[])utf8ToMBS("対象");
				width = 260;
			}
			columns.add(column);
			
			parent = this;
		}
		
		with (m_timer = new Timer) 
		{
			tick 	~= &refreshTick;
			interval = 1_000;
		}
	}
	
	void refreshTick(Object sender, EventArgs ea)
	{
		refreshTask();
	}
	
	
public:
	void refreshTask()
	{
		char[][] taskColumn(char[] type, uint rate, uint ratio, char[] name)
		{
			char[][] row = new char[][4];
			
			row[0] = type;
			
			if (rate == 0) {
				row[1] = std.string.format("%.2f", 0f);
			} else {
				row[1] = std.string.format("%.2f",  cast(double)rate / 1024);
			}
			row[2] = std.string.format("%s%%", ratio);
			
			row[3] = name;
			
			return row;
		}
		
		// クリア
		uint index = m_view.focusedIndex;
		m_view.beginUpdate();
		m_view.clear();
		
		// アップローダ
		s_environ.uploader.signal();
		try {
			foreach (Uploader e; s_environ.uploader.values) {
				m_view.addRow(taskColumn("Uploader", e.rate, e.ratio, e.name));
			}
		} finally {
			s_environ.uploader.release();
		}
		
		// ダウンローダ
		s_environ.downloader.signal();
		try {
			foreach (Downloader e; s_environ.downloader.values) {
				m_view.addRow(taskColumn("Downloader", e.rate, e.ratio, e.name));
			}
		} finally {
			s_environ.downloader.release();
		}
		
		// デコーダ
		s_environ.decoder.signal();
		try {
			foreach (CacheDecoder e; s_environ.decoder.values) {
				m_view.addRow(taskColumn("Decoder", e.rate, e.ratio, e.name));
			}
		} finally {
			s_environ.decoder.release();
		}
		
		// エンコーダ
		s_environ.encoder.signal();
		try {
			foreach (CacheEncoder e; s_environ.encoder.values) {
				m_view.addRow(taskColumn("Encoder", e.rate, e.ratio, e.name));
			}
		} finally {
			s_environ.encoder.release();
		}
		
		m_view.endUpdate();
	}
	
	this(UserInterfaceTask ui)
	{
		super("タスク");
		s_parent 	= ui;
		s_environ	= s_parent.environ;
		initialize();
	}
	
	void startRefreshTimer()
	{
		m_timer.start();
	}
	
	void stopRefreshTimer()
	{
		m_timer.stop();
	}
	
}

class NodeTabPage: TabPage
{
protected:
	Environment	s_environ;
	ListView	m_view;
	Timer		m_timer;
	
	void initialize()
	{
		with(m_view = new ListView)
		{
			ColumnHeader column;
			
			dock 			= DockStyle.FILL;
			font 			= new Font("MS ゴシック", 9f);
			foreColor 		= Color(0);
//			sorting 		= SortOrder.ASCENDING;
			view 			= View.DETAILS;
			fullRowSelect 	= true;
			gridLines 		= true;
			
			with(column = new ColumnHeader)
			{
				column.text = "Type";
				width = 40;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = "Priority";
				width = 55;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = "Speed";
				width = 50;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = "Rate";
				width = 40;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = "Time";
				width = 45;
			}
			columns.add(column);
			
			with(column = new ColumnHeader)
			{
				column.text = "Group(1)";
				width = 80;
			}
			columns.add(column);
			with(column = new ColumnHeader)
			{
				column.text = "Group(2)";
				width = 80;
			}
			columns.add(column);
			with(column = new ColumnHeader)
			{
				column.text = "Group(3)";
				width = 80;
			}
			columns.add(column);
			
			parent = this;
		}
		
		with (m_timer = new Timer) 
		{
			tick 	~= &refreshTick;
			interval = 1_000;
		}
	}
	
	void refreshTick(Object sender, EventArgs ea)
	{
		refreshNode();
//		writef("-refresh tick\n");
	}
	
public:
	void refreshNode()
	{
		char[][] nodeColumn(char[] type, int priority, uint speed,
								uint rate, uint time, char[][] group)
		{
			char[][] row = new char[][8];
			
			row[0] = type;
			row[1] = std.string.toString(priority);
			row[2] = std.string.toString(speed);
			if (rate == 0) {
				row[3] = std.string.format("%.2f", 0f);
			} else {
				row[3] = std.string.format("%.2f",  cast(double)rate / 1024);
			}
			if (time == 0) {
				row[4] = std.string.format("%03d:%02d", 0, 0);
			} else {
				row[4] = std.string.format("%03d:%02d", time / 60, time % 60);
			}
			row[5] = group[0];
			row[6] = group[1];
			row[7] = group[2];
			
			return row;
		}
		
		// クリア
		uint index = m_view.focusedIndex;
		m_view.beginUpdate();
		m_view.clear();
		
//		s_environ.nodeTable.signal();
//		s_environ.upstreamConnection.signal();
//		s_environ.downstreamConnection.signal();
//		s_environ.dlDataConnection.signal();
//		s_environ.ulDataConnection.signal();
		
		// 上流
		try {
			foreach (Connection e; s_environ.upstreamConnection.values(LockerLock.NON_BLOCK).sort) {
//			foreach (Connection e; s_environ.upstreamConnection.values.sort) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				NodeDetails node = e.nodeDetails;
				m_view.addRow(nodeColumn("UP", node.priority, node.speed,
											e.connection.rate,  e.connection.ctl, node.clusteringKey));
			}
			
			// 下流
			foreach (Connection e; s_environ.downstreamConnection.values(LockerLock.NON_BLOCK).sort) {
//			foreach (Connection e; s_environ.downstreamConnection.values.sort) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				NodeDetails node = e.nodeDetails;
				m_view.addRow(nodeColumn("DOWN", node.priority, node.speed,
											e.connection.rate,  e.connection.ctl, node.clusteringKey));
			}
			
			// DLデータ
			foreach (Connection e; s_environ.dlDataConnection.values(LockerLock.NON_BLOCK).sort) {
//			foreach (Connection e; s_environ.dlDataConnection.values.sort) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				NodeDetails node = e.nodeDetails;
				m_view.addRow(nodeColumn("DL", node.priority, node.speed,
											e.connection.rate,  e.connection.ctl, node.clusteringKey));
			}
			
			// ULデータ
			foreach (Connection e; s_environ.ulDataConnection.values(LockerLock.NON_BLOCK).sort) {
//			foreach (Connection e; s_environ.ulDataConnection.values.sort) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				NodeDetails node = e.nodeDetails;
				m_view.addRow(nodeColumn("UL", node.priority, node.speed,
											e.connection.rate,  e.connection.ctl, node.clusteringKey));
			}
			
			// 未接続
			int max = 25; // 25個まで表示
			foreach (NodeDetails e; s_environ.nodeTable.values(LockerLock.NON_BLOCK).sort) {
//			foreach (NodeDetails e; s_environ.nodeTable.values.sort) {
				if (e is null) {
					continue;
				}
				if (max-- <= 0) {
					break;
				}
				m_view.addRow(nodeColumn("WAIT", e.priority, e.speed,
											0, 0, e.clusteringKey));
			}
		} finally {
//			s_environ.upstreamConnection.release();
//			s_environ.downstreamConnection.release();
//			s_environ.dlDataConnection.release();
//			s_environ.ulDataConnection.release();
//			s_environ.nodeTable.release();
		}
		m_view.endUpdate();
	}
	
	void startRefreshTimer()
	{
		m_timer.start();
	}
	
	void stopRefreshTimer()
	{
		m_timer.stop();
	}
	
	this(Environment environ)
	{
		super("ネットワーク");
		initialize();
		s_environ = environ;
	}
}

// 保守画面
class MaintenanceTabPage: TabPage
{
	// 回線速度コンボ用アイテム
	class SpeedItem 
	{
	protected:
		uint	m_speed;
		char[]	m_label;
		
	public:
		this(uint speed, char[] label)
		{
			m_speed = speed;
			m_label = label;
		}
		
		override char[] toString()
		{
			return std.string.format("%s KB/S (%s)", m_speed, m_label);
		}
		
		uint speed()
		{
			return m_speed;
		}
	}
	// GC間隔LV用アイテム
	class IntItem
	{
	protected:
		int	m_int;
		char[] m_string;
	public:
		this(int vint, char[] vstring)
		{
			m_int = vint;
			m_string = vstring;
		}
		
		override char[] toString()
		{
			return m_string;
		}
		
		int toInt()
		{
			return m_int;
		}
	}
	
	// 基本設定
	class BasicConfig
	{
		uint port;
		uint speed;
		char[][3] group;
		char[] downloadFolder;
		char[] uploadFolder;
		char[] cacheFolder;
		
		this(uint port, uint speed, char[] group1, char[] group2, char[] group3, char[] down, char[] up, char[]cache)
		{
			this.port		= port;
			this.speed		= speed;
			this.group[0]	= group1.dup;
			this.group[1]	= group2.dup;
			this.group[2]	= group3.dup;
			downloadFolder	= down;
			uploadFolder	= up;
			cacheFolder		= cache;
		}
	}
	
	// メンバ
	UserInterfaceTask	s_parent;
	Timer				m_timer;
	
	// 基本
	GroupBox		m_basic_box;
	GroupBox		m_initial_box;
	NumericTextBox 	m_port_text;
	ComboBox		m_speed_list;
	TextBox[3]		m_cluster_text;
	Label			m_port_label;
	Label			m_speed_label;
	Label[3]		m_cluster_label;
	Label			m_download_folder_label;
	Label			m_cache_folder_label;
	Label			m_upload_folder_label;
	TextBox			m_download_folder_text;
	TextBox			m_cache_folder_text;
	TextBox			m_upload_folder_text;
	Button			m_download_folder_button;
	Button			m_cache_folder_button;
	Button			m_upload_folder_button;
	ComboBox		m_gc_interval_list;
	
	Button			m_ok, m_undo;
	Label			m_state_label;
	
	Button			m_upnp;
	
	// GCステータス
	GroupBox		m_gc_box;
	GroupBox		m_resource_group;
	Label			m_gc_pool_label;
	Label			m_gc_used_label;
	Label			m_gc_free_label;
	Label			m_gc_pool_value;
	Label			m_gc_used_value;
	Label			m_gc_free_value;
	
	
	// 初期ノード
	Label			m_self_label;
	TextBox			m_self_text;
	Label			m_url_label;
	TextBox			m_url_text;
	Button			m_node_get;
	
	// POENYステータス
	GroupBox		m_state_box;
	Label			m_rate_label;
	Label			m_rate_value;
	Label			m_keys_label;
	Label			m_keys_value;
	Label			m_nodes_label;
	Label			m_nodes_value;
	Label			m_download_label;
	Label			m_download_value;
	
	// 初期化
	void initialize()
	{
		// 基本設定
		with (m_basic_box = new GroupBox)
		{
			text		= "基本設定";
			bounds		= Rect(10, 5, 300, 320);
//			backColor 	= Color(200, 200, 200);
			parent		= this;
		}
		
		with (m_port_label = new Label)
		{
			text			= "ポート";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 15, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_port_text = new NumericTextBox)
		{
			maxLength		= 5;
			bounds			= Rect(80, 15, 50, 20);
			parent			= m_basic_box;
		}
		
		with (m_upnp = new Button)
		{
			text			= "Open(UPnP)";
			bounds			= Rect(140, 15, 120, 20);
			parent			= m_basic_box;
			click			~= &clickOpenUPnP;
			enabled			= false;
		}
		
		with (m_speed_label = new Label)
		{
			text			= "回線速度";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 40, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_speed_list = new ComboBox)
		{
			text			= "DEFAULT";
			bounds			= Rect(80, 40, 180, 20);
			parent			= m_basic_box;
			dropDownStyle	= ComboBoxStyle.DROP_DOWN_LIST;
			items.add 		= new SpeedItem(1000, "T3, FTTH 10M");
			items.add		= new SpeedItem(500, "ADSL 50M");
			items.add		= new SpeedItem(120, "ADSL 1.5M-24M");
			items.add		= new SpeedItem(50, "CATV");
			items.add		= new SpeedItem(15, "PHS 128Kbps");
		}
		
		with (m_cluster_label[0] = new Label)
		{
			text			= "グループ(1)";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 65, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_cluster_text[0] = new TextBox)
		{
			bounds			= Rect(80, 65, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		
		with (m_cluster_label[1] = new Label)
		{
			text			= "グループ(2)";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 90, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_cluster_text[1] = new TextBox)
		{
			bounds			= Rect(80, 90, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		with (m_cluster_label[2] = new Label)
		{
			text			= "グループ(3)";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 115, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_cluster_text[2] = new TextBox)
		{
			bounds			= Rect(80, 115, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		
		with (m_download_folder_label = new Label)
		{
			text			= "Down DIR";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 140, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_download_folder_text = new TextBox)
		{
			bounds			= Rect(80, 140, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		
		with (m_download_folder_button = new Button)
		{
			text			= "..";
			bounds			= Rect(270, 140, 20, 20);
			click			~= &clickDownload;
			parent			= m_basic_box;
		}
		
		with (m_upload_folder_label = new Label)
		{
			text			= "Up DIR";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 165, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_upload_folder_text = new TextBox)
		{
			bounds			= Rect(80, 165, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		
		with (m_upload_folder_button = new Button)
		{
			text			= "..";
			bounds			= Rect(270, 165, 20, 20);
			click			~= &clickUpload;
			parent			= m_basic_box;
		}
		
		with (m_cache_folder_label = new Label)
		{
			text			= "Cache DIR";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 190, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_cache_folder_text = new TextBox)
		{
			bounds			= Rect(80, 190, 180, 20);
			maxLength		= 255;
			parent			= m_basic_box;
		}
		
		with (m_cache_folder_button = new Button)
		{
			text			= "..";
			bounds			= Rect(270, 190, 20, 20);
			click			~= &clickCache;
			parent			= m_basic_box;
		}
/*		
		with (new Label)
		{
			text			= "GC MT LV";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 215, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_basic_box;
		}
		
		with (m_gc_interval_list = new ComboBox)
		{
			text			= "DEFAULT";
			bounds			= Rect(80, 215, 180, 20);
			parent			= m_basic_box;
			dropDownStyle	= ComboBoxStyle.DROP_DOWN_LIST;
			items.add 		= new IntItem(1, cast(char[])utf8ToMBS("0 (明示的回収をしない)"));
			items.add 		= new IntItem(1, cast(char[])utf8ToMBS("1 (超低速回線)"));
			items.add		= new IntItem(2, cast(char[])utf8ToMBS("2 (標準)"));
			items.add		= new IntItem(3, cast(char[])utf8ToMBS("3 (超高速回線)"));
		}
*/		
		with (m_ok = new Button)
		{
			text			= "設定";
			bounds			= Rect(10, 290, 50, 20);
			parent			= m_basic_box;
			click			~= &clickSet;
		}
		
		with (m_undo = new Button)
		{
			text			= "リセット";
			bounds			= Rect(70, 290, 50, 20);
			parent			= m_basic_box;
			
			click			~= &clickReset;
		}
		
		with (m_state_label = new Label)
		{
			text			= "";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			backColor		= Color(200, 200, 200);
			bounds			= Rect(130, 290, 130, 20);
			parent			= m_basic_box;
		}
		
		// GC
		with (m_gc_box = new GroupBox)
		{
			text		= "GC";
			bounds		= Rect(320, 5, 155, 104);
//			backColor 	= Color(200, 200, 200);
			parent		= this;
		}
		
		with (m_gc_pool_label = new Label)
		{
			text			= "合計";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 15, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		with (m_gc_pool_value = new Label)
		{
			text			= "99,999,999 B";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 15, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		with (m_gc_free_label = new Label)
		{
			text			= "利用可能";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 40, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		with (m_gc_free_value = new Label)
		{
			text			= "99,999,999 B";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 40, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		with (m_gc_used_label = new Label)
		{
			text			= "利用中";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 65, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		with (m_gc_used_value = new Label)
		{
			text			= "99,999,999 B";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 65, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_gc_box;
		}
		
		// 資源
		with (m_state_box = new GroupBox)
		{
			text		= "ポエニーステータス";
			bounds		= Rect(320, 115, 155, 305);
//			backColor 	= Color(200, 200, 200);
			parent		= this;
		}
		
		with (m_rate_label = new Label)
		{
			text			= "通信速度";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 15, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_rate_value = new Label)
		{
			text			= "999,999KB";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 15, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_keys_label = new Label)
		{
			text			= "キー数";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 40, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_keys_value = new Label)
		{
			text			= "99,999  ";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 40, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_nodes_label = new Label)
		{
			text			= "ノード数";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 65, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_nodes_value = new Label)
		{
			text			= "99,999  ";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 65, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_download_label = new Label)
		{
			text			= "DL予約数";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 90, 50, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		with (m_download_value = new Label)
		{
			text			= "99,999  ";
			textAlign		= ContentAlignment.MIDDLE_RIGHT;
			bounds			= Rect(70, 90, 75, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_state_box;
		}
		
		// 初期ノード
		with (m_initial_box = new GroupBox)
		{
			text		= "初期ノード";
			bounds		= Rect(10, 330, 300, 90);
//			backColor 	= Color(200, 200, 200);
			parent		= this;
		}
		
		with (m_self_label = new Label)
		{
			text			= "自局ノード";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 15, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_initial_box;
		}
		
		with (m_self_text = new TextBox)
		{
			readOnly		= true;
			bounds			= Rect(80, 15, 180, 20);
			parent			= m_initial_box;
		}
		
		with (m_url_label = new Label)
		{
			text			= "URI";
			textAlign		= ContentAlignment.MIDDLE_LEFT;
			bounds			= Rect(10, 40, 60, 20);
			backColor		= Color(200, 200, 200);
			parent			= m_initial_box;
		}
		
		with (m_url_text = new TextBox)
		{
			bounds			= Rect(80, 40, 180, 20);
			maxLength		= 512;
			parent			= m_initial_box;
			enabled			= false;
		}
		
		with (m_node_get = new Button)
		{
			text			= "取得";
			bounds			= Rect(10, 65, 50, 20);
			parent			= m_initial_box;
			
			click			~= &clickGetInitialNodes;
			enabled			= false;
		}
		
		with (m_timer = new Timer) 
		{
			tick 	~= &refreshTick;
			interval = 5_000;
		}
		
		readBasicConfig();
		refreshState();
		setSelfAddress();
	}
	
	void clickDownload(Object sender, EventArgs ea)
	{
		selectDir(m_download_folder_text);
	}
	
	void clickUpload(Object sender, EventArgs ea)
	{
		selectDir(m_upload_folder_text);
	}
	
	void clickCache(Object sender, EventArgs ea)
	{
		selectDir(m_cache_folder_text);
	}
	
	void selectDir(TextBox textbox)
	{
		FolderBrowserDialog dialog = new FolderBrowserDialog;
		
		dialog.selectedPath(textbox.text);
		
		if (dialog.showDialog == DialogResult.OK) {
			textbox.text = dialog.selectedPath;
		}
	}
	
	void refreshTick(Object sender, EventArgs ea)
	{
		refreshState();
//		writef("-refresh tick state\n");
	}
	
	void startTimer()
	{
		m_timer.start();
	}
	
	void stopTimer()
	{
		m_timer.stop();
	}
	
	void clickOpenUPnP(Object sender, EventArgs ea)
	{
		notImplementedDialog();
	}
	
	void clickSet(Object sender, EventArgs ea)
	{
		BasicConfig basic = getBasicConfig();
		
		if (checkBasicConfig(basic)) {
			m_state_label.text = "設定中…";
			m_state_label.update;
			m_ok.enabled 	= false;
			m_undo.enabled	= false;
			try {
				writeBasicConfig(getBasicConfig);
			} finally {
				m_ok.enabled 	= true;
				m_undo.enabled	= true;
			}
		} else {
			m_state_label.text = "設定失敗!!";
		}
	}
	
	void clickReset(Object sender, EventArgs ea)
	{
		readBasicConfig();
	}
	
	// 基本設定取得
	BasicConfig getBasicConfig()
	{
		SpeedItem	speed_item;
		uint port;
		
		if (m_port_text.text.length != 0) {
			port 	= toUint(m_port_text.text);
		} else {
			port	= 0;
		}
		speed_item 	= cast(SpeedItem)m_speed_list.selectedItem;
		
		return new BasicConfig(port, speed_item.speed, 
					m_cluster_text[0].text, m_cluster_text[1].text, m_cluster_text[2].text,
					m_download_folder_text.text, m_upload_folder_text.text, m_cache_folder_text.text);
	}
	
	// 基本設定設定
	void setBasicConfig(BasicConfig config)
	in {
		assert(config.group.length == m_cluster_text.length);
	}
	body {
		int index;
		m_port_text.text = std.string.toString(config.port);
		
		for (int i = 0; i < m_speed_list.items.count; ++i) {
			if ((cast(SpeedItem)m_speed_list.items[i]).speed <= config.speed) {
				index = m_speed_list.items.indexOf(m_speed_list.items[i]);
				break;
			}
		}
		m_speed_list.selectedIndex = index;
		
		for (int i = 0; i < config.group.length; ++i)
		{
			m_cluster_text[i].text = config.group[i];
		}
		
		m_download_folder_text.text		= config.downloadFolder;
		m_upload_folder_text.text		= config.uploadFolder;
		m_cache_folder_text.text		= config.cacheFolder;
	}
	
	// 基本設定読み込み
	public void readBasicConfig()
	{
		Configuration 	config 	= s_parent.environ.configuration;
		BasicConfig 	basic;
		
		basic = new BasicConfig(config.fileSharingPort, config.speed,
					config.clusteringKey[0], config.clusteringKey[1], config.clusteringKey[2],
					config.downloadFolder, config.uploadFolder, config.cacheFolder);
		
		setBasicConfig(basic);
		m_state_label.text = "";
		
		m_port_text.focus();
	}
	
	// 基本設定値チェック
	bool checkBasicConfig(BasicConfig config)
	{
		
		try {
			if (!std.file.isdir(m_download_folder_text.text)) {
				msgBox("Downloadフォルダが不正です", "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
			}
		} catch (Exception e) {
			msgBox("Downloadフォルダが不正です\n" ~ e.toString, "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
		}
		try {
			if (!std.file.isdir(m_upload_folder_text.text)) {
				msgBox("Uploadフォルダが不正です", "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
			}
		} catch (Exception e) {
			msgBox("Uploadフォルダが不正です\n" ~ e.toString, "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
		}
		try {
			if (!std.file.isdir(m_cache_folder_text.text)) {
				msgBox("Cacheフォルダが不正です", "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
			}
		} catch (Exception e) {
			msgBox("Cacheフォルダが不正です\n" ~ e.toString, "Error",
					MsgBoxButtons.OK, MsgBoxIcon.ERROR);
				return false;
		}
		
		return true;
	}
	
	// 基本設定書き込み
	void writeBasicConfig(BasicConfig basic)
	{
		Configuration config = s_parent.environ.configuration;
		char[] down_back, up_back, cache_back;
		
		bool node_reboot = false, disconnect = false, task_reboot = false;
		uint backup_port = config.fileSharingPort;
		
		scope(success) m_state_label.text = "設定成功";
		scope(failure) m_state_label.text = "設定失敗!!";
		
		if (config.fileSharingPort != basic.port) {
			node_reboot = true;
		}
		if (config.speed != basic.speed) {
			disconnect = true;
		}
		
		if (config.cacheFolder != basic.cacheFolder
			|| config.downloadFolder != basic.downloadFolder
			|| config.uploadFolder != basic.uploadFolder) 
		{
			task_reboot = true;
		}
		
		
		config.fileSharingPort	= basic.port;
		config.speed 			= basic.speed;
		config.clusteringKey	= basic.group;
		
		if (node_reboot || disconnect) {
			if (!s_parent.rebootNode) {
				config.fileSharingPort = backup_port;
				m_state_label.text = "設定失敗!!";
			}
		}
		
		// ここ排他しないと･･
		down_back 				= config.downloadFolder.dup;
		up_back 				= config.uploadFolder.dup;
		cache_back 				= config.cacheFolder.dup;
		config.downloadFolder	= basic.downloadFolder;
		config.uploadFolder		= basic.uploadFolder;
		config.cacheFolder		= basic.cacheFolder;
		s_parent.saveConfiguration();
		config.downloadFolder	= down_back;
		config.uploadFolder		= up_back;
		config.cacheFolder		= cache_back;
		
		setSelfAddress(); // アドレス更新
		
		if (task_reboot) {
			msgBox("変更フォルダは再起動後に有効になります\n", "フォルダが変更されました",
					MsgBoxButtons.OK, MsgBoxIcon.INFORMATION);
		}
	}
	
	// 自局アドレス設定
	void setSelfAddress()
	{
		InternetAddress	self 	= getSystemIPAddress(s_parent.environ.configuration.fileSharingPort);
		uint 			test_ip	= self.addr;
		
		// ある程度の微妙なテスト
		if ((test_ip & 0xffff0000) == 0xC0A80000	// 192.168.*.*
			|| test_ip == 0x7F000001)				// 127.0.0.1
		{
			m_self_text.text = "ER: ローカルアドレス";
		} else {
			NyNodeAddress address = new NyNodeAddress;
			
			m_self_text.text = address.toHashString(self);
		}
	}
	
	void refreshState()
	{
		GCStats gc;
		uint	rate = 0;
		uint	dl_count = 0;
		
		// 通信レート合計
		foreach (Connection e; s_parent.environ.upstreamConnection.values(LockerLock.NON_BLOCK)) {
//		foreach (Connection e; s_parent.environ.upstreamConnection.values) {
			if (e is null || e.connection is null) {
				continue;
			}
			rate += e.connection.rate;
		}
		foreach (Connection e; s_parent.environ.downstreamConnection.values(LockerLock.NON_BLOCK)) {
//		foreach (Connection e; s_parent.environ.downstreamConnection.values) {
			if (e is null || e.connection is null) {
				continue;
			}
			rate += e.connection.rate;
		}
		foreach (Connection e; s_parent.environ.ulDataConnection.values(LockerLock.NON_BLOCK)) {
//		foreach (Connection e; s_parent.environ.ulDataConnection.values) {
			if (e is null || e.connection is null) {
				continue;
			}
			rate += e.connection.rate;
		}
		foreach (Connection e; s_parent.environ.dlDataConnection.values(LockerLock.NON_BLOCK)) {
//		foreach (Connection e; s_parent.environ.dlDataConnection.values) {
			if (e is null || e.connection is null) {
				continue;
			}
			rate += e.connection.rate;
		}
		
		// GCステート
		std.gc.getStats(gc);
		m_gc_pool_value.text = comma(gc.poolsize) ~ " B";
		m_gc_free_value.text = comma(gc.freelistsize) ~ " B";
		m_gc_used_value.text = comma(gc.poolsize - gc.freelistsize) ~ " B";
		
		// ポエニーステート
		m_rate_value.text	= format("%.2f", cast(double)rate / 1024) ~ " KB/S";
		m_keys_value.text	= comma(s_parent.environ.keyTable.length) ~ "  ";
		m_nodes_value.text	= comma(s_parent.environ.nodeTable.length) ~ "  ";
		
		s_parent.environ.downloadList.signal();
		try {
			foreach (DownloadItem item; s_parent.environ.downloadList.values) {
				if (! item.isComplete) {
					++dl_count; // DL試行中
				}
			}
		} finally {
			s_parent.environ.downloadList.release();
		}
		m_download_value.text	= comma(dl_count) ~ "  ";
		
		m_gc_pool_value.update;
		m_gc_free_value.update;
		m_gc_used_value.update;
		m_rate_value.update;
		m_keys_value.update;
		m_nodes_value.update;
		m_download_value.update;
	}
	
	// 初期ノード取得
	void clickGetInitialNodes(Object sender, EventArgs ea)
	{
		notImplementedDialog();
	}
	
	this(UserInterfaceTask parent)
	{
		super("コックピット");
		s_parent	= parent;
		
		initialize();
	}
}

/* ポエニーメイン画面	*/
class PoenyMainForm: Form
{
protected:
	UserInterfaceTask	s_parent;
	
	TabControl		 	m_tab;
	TabPage				m_poem, m_cache;
	SearchTabPage		m_search;
	NodeTabPage			m_node;
	TaskTabPage			m_task;
	MaintenanceTabPage	m_maintenance;
	DownloadTabPage		m_download;
	bool				m_task_timer, m_node_timer, m_maintenance_timer;
	
	void drawMenu(Form form)
	{
		
	}
	
	TabPage drawSearchPage()
	{
		m_search = new SearchTabPage(s_parent);
		return m_search;
	}
	
	TabPage drawTaskPage()
	{
		m_task	= new TaskTabPage(s_parent);
		return m_task;
	}
	
	TabPage drawPoemPage()
	{
		m_poem = new PoemTabPage(s_parent.environ);
		return m_poem;
	}
	
	TabPage drawNodePage()
	{
		m_node = new NodeTabPage(s_parent.environ);
		return m_node;
	}
	
	TabPage drawSettingPage()
	{
		m_maintenance = new MaintenanceTabPage(s_parent);
		return m_maintenance;
	}
	
	TabPage drawDownloadPage()
	{
		m_download = new DownloadTabPage(s_parent);
		return m_download;
	}
	
	TabPage drawCachePage()
	{
		m_cache	= new CacheTabPage(s_parent);
		return m_cache;
	}
	
	void drawMainTab()
	{
		m_tab = new TabControl;
		
		with (m_tab)
		{
			dock 			= DockStyle.FILL;
			alignment		= TabAlignment.TOP;
			appearance		= TabAppearance.NORMAL;
//			size			= Size(495, 450);
			
			tabPages.add(drawSearchPage());
			tabPages.add(drawPoemPage());
			tabPages.add(drawDownloadPage());
			tabPages.add(drawTaskPage());
			tabPages.add(drawNodePage());
			tabPages.add(drawCachePage());
			tabPages.add(drawSettingPage());
			
			parent = this;
		}
	}
	
	void selectedIndexChangedHandle(Object sender, EventArgs ea)
	{
		switch (m_tab.selectedIndex) {
		case 0:	// 検索
			m_search.setFocus();
			break;
		case 1:	// ポエムを書く
			break;
		case 2: // ダウンロード
			m_download.refreshItem();
			break;
		case 3:	// タスク
			m_task.startRefreshTimer();
			m_task.refreshTask();
			m_task_timer = true;
			break;
		case 4:	// ノード
			m_node.startRefreshTimer();
			m_node.refreshNode();
			m_node_timer = true;
			break;
		case 5: // キャッシュ
			break;
		case 6:	// 保守
			m_maintenance.readBasicConfig();
			m_maintenance.refreshState();
			m_maintenance.startTimer();
			m_maintenance_timer = true;
			break;
		default:
			break;
		}
	}
	
	void selectedIndexChangingHandle(Object sender, CancelEventArgs ea)
	{
		switch (m_tab.selectedIndex) {
		case 0:	// 検索
		case 1:	// ポエム
		case 2:	// ダウンロード
			break;
		case 3: // タスク
			m_task.stopRefreshTimer();
			m_task_timer = false;
			break;
		case 4:	// ネットワーク
			m_node.stopRefreshTimer();
			m_node_timer = false;
			break;
		case 5: // キャッシュ
			break;
		case 6:	// 保守
			m_maintenance.stopTimer();
			m_maintenance_timer = false;
			break;
		default:
			break;
		}
	}
	
	void initialize()
	{
		// menu
		drawMenu(this);
		
		// tabs
		drawMainTab();
		
		// ハンドラ
		mainTab.selectedIndexChanged ~= &selectedIndexChangedHandle;
		mainTab.selectedIndexChanging ~= &selectedIndexChangingHandle;
	}
	
	TabControl mainTab()
	{
		return m_tab;
	}
	
	this(UserInterfaceTask parent)
	{
		s_parent 		= parent;
		
		text 			= MAIN_FORM_TITLE;
		dockPadding.all = 0;
		size			= Size(500, 480);
		formBorderStyle = FormBorderStyle.SIZABLE;
		
		initialize();
	}
}

/* PoenTropyメイン画面	*/
class PoenTropyMainForm: PoenyMainForm
{
	this(UserInterfaceTask uitask)
	{
		super(uitask);
	}
	
	override void drawMainTab()
	{
		m_tab = new TabControl;
		
		with (m_tab)
		{
			dock 			= DockStyle.FILL;
			alignment		= TabAlignment.TOP;
			appearance		= TabAppearance.NORMAL;
			
//			tabPages.add(drawDownloadPage());
			tabPages.add(drawPoemPage());
			tabPages.add(drawNodePage());
			tabPages.add(drawSettingPage());
			parent = this;
		}
	}
	
	override void selectedIndexChangedHandle(Object sender, EventArgs ea)
	{
		switch (m_tab.selectedIndex) {
		case 0:	// ポエム
			break;
		case 1:	// ネットワーク
			m_node.startRefreshTimer();
			m_node.refreshNode();
			m_node_timer = true;
			break;
		case 2:	// 保守
			m_maintenance.readBasicConfig();
			m_maintenance.refreshState();
			m_maintenance.startTimer();
			m_maintenance_timer = true;
			break;
		default:
			break;
		}
	}
	
	override void selectedIndexChangingHandle(Object sender, CancelEventArgs ea)
	{
		switch (m_tab.selectedIndex) {
		case 0:	// ポエム
			break;
		case 1:	// ネットワーク
			m_node.stopRefreshTimer();
			m_node_timer = false;
			break;
		case 2:	// 保守
			m_maintenance.stopTimer();
			m_maintenance_timer = false;
			break;
		default:
			break;
		}
	}
	
}
