﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module task.keytask;

// キー管理タスク

private import 	std.thread,
				std.c.time,
				std.utf,
				std.random;

private import 	pes.net.winny.key,
				pes.arraybox,
				pes.system.all,
				pes.wstring,
				pes.random,
				pes.variant;

private import task.uicommand;
private import task.timertask;
private import common.task;
private import common.environment;
private import common.keytable;

debug(key_task)
{
	import pes.prompt;
}

class KeyTask: TaskBase
{
protected:
	const uint DELETE_KEY_TICK 	= 0;
	const uint LOAD_CACHE_TICK	= 1;
	
	override TaskId taskId()
	{
		return TaskId.KEY;
	}
	
	override void initialize()
	{
		gcLimit = gcLimit / 8;
		//setPriority(PRIORITY.INCREASE);
		environ.keyTable.clear();
		refreshLocalKeys();
	}
	
	override void terminate()
	{
		// nothing
	}
	
	override ReturnCode control()
	{
		ReturnCode 	result;
		
		/* キータスクメインループ */
		while (true) {
			result = control_x();
			if (result != ReturnCode.SUCCESS) {
				break;
			}
			delayTask(10);
		}
		
		return result;
	}
	
	ReturnCode control_x()
	{
		KeyInformation[] keys;
		Mail mail = waitForRequest();
		
		switch (mail.messageId) {
		case MessageId.TERMINATE:
			return ReturnCode.TERMINATE;	/* 終了	*/
			
		case MessageId.CLOCK_TICK:
			switch ((cast(VLong)mail.data).toLong) {
			case DELETE_KEY_TICK:
				environ.keyTable.tick30();
				break;
				
			case LOAD_CACHE_TICK:
				refreshLocalKeys();
				break;
			default:
				assert(0);
			}
			break;
			
		case MessageId.ADD_KEY:
			// キー追加
			// クエリタスクと同じタイミングで動かないように
			// ここでディレイを入れる
			delayTask(30);
			addKey((cast(KeyInformationArray)mail.data).array);
			break;
			
		default:
			break;
		}
		
		fullCollect();
		
		return ReturnCode.SUCCESS;
	}
	
	void refreshLocalKeys()
	{
		NyKeyInformation[] keys =  environ.cacheHeader.completeCacheKeys(config);
		addKey(keys);
	}
	
	void addKey(NyKeyInformation[] keyinfo)
	{
		uint 			max = config.keyMax;
		KeyInformation 	key;
		
		if (keyinfo.length == 0) {
			return;
		}
		environ.keyTable.signal();
		try {
			environ.keyTable.addKeyArray(keyinfo);
			foreach (inout NyKeyInformation e; keyinfo) {
				e.destroy();
				delete e;
				e = null;
			}
			debug(key_task) mbswritef(derr, "Key: キー追加(数:%s, :合計:%s/%s)\n",
											keyinfo.length, environ.keyTable.length, max);
			
			delete keyinfo;
			keyinfo = null;
			if (environ.keyTable.length > max) {
				log("キー限界");
				environ.keyTable.clear(max / 2);
				log("キークリア");
			}
		} finally {
			environ.keyTable.release();
		}
	}
	
public:
	this(Environment environ, TaskResource resource)
	{
		super(MailId.KEY_REQUEST,
				MailId.KEY_RESPONSE, environ, resource);
		
		// キー削除タイマー
		setMessageTimer(this, 30, MessageTimer.TTL.ANY, new VLong(DELETE_KEY_TICK));
		
		// キャッシュロードタイマー
		setMessageTimer(this, 60, MessageTimer.TTL.ANY, new VLong(LOAD_CACHE_TICK));
	}
}

