﻿/**
 * Authors: http://d.hatena.ne.jp/ultraist/
 */
module uitask.uitask;
/*
 * UI タスクのソケットインターフェースバージョン
 *
 * 注意!: 途中からGUIにしたので、このコードは古いです。
 *        Linux用にそのうちきちんと作るかもしれません。
 * 
 */
 
private import	std.stdio,
				std.thread,
				std.string,
				std.socket,
				std.c.time,
				std.gc;

private import	pes.system.all,
 				pes.arraybox,
 				pes.conv,
				pes.net.winny.all;

private import 	common.task,
				common.environment,
				uitask.command,
				cachetask.cachetask,
				nodetask.nodetask,
				keytask.keytask,
				querytask.querytask,
				controltask.controltask,
				linktask.linktask,
				connectiontask.connectiontask;

debug(ui_task)
{
	import pes.prompt;
}

class UserInterfaceTask: TaskBase
{
protected:
	Socket			m_socket;
	uint			m_port;
	uint			m_ny_default_port;
	
	NodeTask		m_node_task;
	KeyTask			m_key_task;
	QueryTask		m_query_task;
	ControlTask		m_control_task;
	CacheTask		m_cache_task;
	LinkTask		m_link_task;
	ConnectionTask	m_connection_task;
	
	
	/* UIメインループ */
	override int mainLoop()
	{
		ReturnCode result;
		
		m_thread = Thread.getThis();
		
		initialize();
		
	_MAIN_LOOP:
		while (true)
		{
			try {
				result = control();
			} catch (Exception e) {
				log("control 例外: " ~ e.toString);
				result = ReturnCode.FAILURE;
			} finally {
				terminate();
			}
			
			switch (result) {
			case ReturnCode.TERMINATE:
				log("Terminate");
				result = ReturnCode.SUCCESS;
				break _MAIN_LOOP;
			
			case ReturnCode.CONTINUE:
				log("Continue");
				try {
					initialize();
				} catch (Exception e) {
					log("initialize 例外: " ~ e.toString);
					result = ReturnCode.FAILURE;
					break _MAIN_LOOP;
				}
				continue _MAIN_LOOP;
				
			case ReturnCode.FAILURE:
				break _MAIN_LOOP;
			// default: exception
			}
		}
		
		return cast(int)result;
	}
	
	override TaskId taskId()
	{
		return TaskId.UI;
	}
	
	override void initialize()
	{
		int optval = 1;
		
		
		m_socket = new TcpSocket();
		m_socket.setOption(SocketOptionLevel.SOCKET, SocketOption.REUSEADDR, optval);
		m_socket.blocking(true);
		m_socket.bind(new InternetAddress(m_port));
		m_socket.listen(1);	// 今は一人用
	}
	
	override void terminate()
	{
		if (m_socket !is null) {
			m_socket.close();
		}
		endTask();
	}
	
	override ReturnCode control()
	{
		Socket client = m_socket.accept();
		m_socket.close();
		
		if (!startTask()) {
			throw new Error("タスクの起動に失敗:再チャレンジ求む");
		}
		
		uiControl(client);
		
		return ReturnCode.CONTINUE;
	}
	
	void uiControl(Socket client)
	{
		debug(ui_task) mbswritef(derr, "UI: 接続\n");
		SocketSet 			sockset 		= new SocketSet;
		timeval				timeup;
		int 				recvlen, result;
		ubyte 				buff[4096];
		
		
		client.blocking = false;
		while (true)
		{
			try {
				timeup.seconds = 10;	// 10秒
				
				sockset.reset();
				sockset.add(client);
				
				result = Socket.select(sockset, null, null, &timeup);
				if (result == 0) {
					VERY_UGLY_FUNCTION();
					continue;
				} else if (! sockset.isSet(client)) {
					client.close();
					return;
				}
				
				recvlen = client.receive(buff);
				if (recvlen <= 0) {
					client.close();
					debug(ui_task) mbswritef(derr, "UI: 切断\n");
					return;
				}
			} catch {
				client.close();
				return;
			}
			
			uiControl_x(client, buff[0 .. recvlen]);
			delayTask(5);
		}
	}
	
	void uiControl_x(Socket client, ubyte[] buff)
	{
		UIRequestCommand 	req_command		= new UIRequestCommand;
		UIResponseCommand 	error_command	= new UIResponseCommand;
		UIResponseCommand	res_command;
		ubyte[]				packet;
		error_command.status = RequestStatus.NG;
		
		
		try {
			debug(ui_task) mbswritef(derr, "UI: 受信(%s)\n", buff);
			
			req_command.unpack(buff);
			switch (req_command.commandCode) {
			case UICommandCode.QSET_LISTENER_PORT:
				res_command = setListenerPort(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QGET_LISTENER_PORT:
				res_command = getListenerPort(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QSET_SPEED:
				res_command = setSpeed(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QGET_SPEED:
				res_command = getSpeed(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QGET_CLUSTERING_KEY:
				res_command = getClusteringKey(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QSET_CLUSTERING_KEY:
				res_command = setClusteringKey(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QADD_SEARCH_WORD:
				res_command = addSearchWord(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QREMOVE_SEARCH_WORD:
				res_command = removeSearchWord(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QSEARCH:
				res_command = searchFile(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			case UICommandCode.QGET_NODE:
				res_command = getNodeInformation(req_command);
				packet		= res_command.pack;
				client.send(packet);
				break;
			default:
				break;
			}
			res_command = null;
			//delete packet,
			packet = null;
			
		} catch (Exception e) {
			debug(ui_task) mbswritef(derr, "UI: 例外(%s)\n", e.toString);
			error_command.commandCode = req_command.commandCode;
			client.send(error_command.pack);
		}
		
		std.gc.fullCollect(); // ここは本物
	}
	
	
	// リスナーポート設定
	UIResponseCommand setListenerPort(UIRequestCommand req_command)
	{
		UIResponseCommand 	res_command	= new UIRSetListenerPort;
		UIQSetListenerPort	port 		= new UIQSetListenerPort(req_command);
		Mail mail;
		ushort				backup_port	= config.fileSharingPort;
		
		debug(ui_task) mbswritef(derr, "UI: リスナーポート設定(port:%d)\n", port.port);
		config.fileSharingPort = port.port;
		
		// ノードタスクへ再起動要求 
		endEvent(EventId.TASK_START, TaskId.NODE);
		sendRequestTo(MailId.NODE_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
		mail = waitForResponse(); // 同期通信
		
		sendRequestTo(MailId.CONNECTION_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
		
		
		if (mail.status == RequestStatus.OK) {
			// OK
			endEvent(EventId.TASK_START, TaskId.NODE);
			sendRequestTo(MailId.CONNECTION_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
			waitForStartEvent(EventId.TASK_START, TaskId.CONNECTION);
			waitForStartEvent(EventId.TASK_START, TaskId.NODE);
			
			res_command.status = UIStatus.OK;
		} else {
			// NG
			startEvent(EventId.TASK_START, TaskId.NODE);
			config.fileSharingPort = backup_port;
			res_command.status = UIStatus.NG;
		}
		
		return res_command;
	}
	
	// リスナーポート取得
	UIResponseCommand getListenerPort(UIRequestCommand req_command)
	{
		UIRGetListenerPort 	res_command	= new UIRGetListenerPort;
		UIQGetListenerPort	port 		= new UIQGetListenerPort(req_command);
		Mail mail;
		
		debug(ui_task) mbswritef(derr, "UI: リスナーポート取得\n");
		
		res_command.port	= config.fileSharingPort;
		res_command.status	= UIStatus.OK;
		
		return res_command;
	}
	
	
	// 回線速度取得
	UIResponseCommand getSpeed(UIRequestCommand req_command)
	{
		UIRGetSpeed		 	res_command	= new UIRGetSpeed;
		UIQGetSpeed			speed 		= new UIQGetSpeed(req_command);
		
		debug(ui_task) mbswritef(derr, "UI: 回線速度取得\n");
		res_command.speed = config.speed;
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	
	// 回線速度設定
	UIResponseCommand setSpeed(UIRequestCommand req_command)
	{
		UIResponseCommand 	res_command	= new UIRSetSpeed;
		UIQSetSpeed			speed 		= new UIQSetSpeed(req_command);
		Mail mail;
		
		debug(ui_task) mbswritef(derr, "UI: 回線速度設定(speed:%d)\n", speed.speed);
		config.speed = speed.speed;
		
		endEvent(EventId.TASK_START, TaskId.CONNECTION);
		sendRequestTo(MailId.CONNECTION_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
		sendRequestTo(MailId.NODE_REQUEST, Mail.makeRequest(MessageId.REFRESH_NODE, null));
		waitForStartEvent(EventId.TASK_START, TaskId.CONNECTION);
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	
	// クラスタリングキー取得
	UIResponseCommand getClusteringKey(UIRequestCommand req_command)
	{
		UIRGetClusteringKey	res_command	= new UIRGetClusteringKey;
		UIQGetClusteringKey	speed 		= new UIQGetClusteringKey(req_command);
		
		debug(ui_task) mbswritef(derr, "UI: クラスタリングキー取得\n");
		res_command.clusteringKey(config.clusteringKey);
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	// クラスタリングキー設定
	UIResponseCommand setClusteringKey(UIRequestCommand req_command)
	{
		UIRSetClusteringKey 	res_command	= new UIRSetClusteringKey;
		UIQSetClusteringKey		key 		= new UIQSetClusteringKey(req_command);
		Mail mail;
		
		debug(ui_task) mbswritef(derr, "UI: クラスタリングキー設定 %s\n", key.clusteringKey(0));
		debug(ui_task) mbswritef(derr, "UI: クラスタリングキー設定 %s\n", key.clusteringKey(1));
		debug(ui_task) mbswritef(derr, "UI: クラスタリングキー設定 %s\n", key.clusteringKey(2));
		
		config.clusteringKey(key.clusteringKey);
		
		// コネクションとノードを再起動
		endEvent(EventId.TASK_START, TaskId.CONNECTION);
		
		sendRequestTo(MailId.NODE_REQUEST, Mail.makeRequest(MessageId.REFRESH_NODE, null));
		sendRequestTo(MailId.CONNECTION_REQUEST, Mail.makeRequest(MessageId.REBOOT, null));
		
		waitForStartEvent(EventId.TASK_START, TaskId.CONNECTION);
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	// 検索ワード追加
	UIResponseCommand addSearchWord(UIRequestCommand req_command)
	{
		UIResponseCommand 	res_command	= new UIRAddSearchWord;
		UIQAddSearchWord	word 		= new UIQAddSearchWord(req_command);
		
		debug(ui_task) mbswritef(derr, "UI: 検索登録(word:%s)\n", word.word);
		
		sendRequestTo(MailId.QUERY_REQUEST, 
				Mail.makeRequest(MessageId.ADD_SEARCH_WORD, new ArrayBox!(char)(word.word)));
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	// 検索ワード削除
	UIResponseCommand removeSearchWord(UIRequestCommand req_command)
	{
		UIResponseCommand 	res_command	= new UIRAddSearchWord;
		UIQRemoveSearchWord	word 		= new UIQRemoveSearchWord(req_command);
		
		debug(ui_task) mbswritef(derr, "UI: 検索削除(word:%s)\n", word.word);
		
		sendRequestTo(MailId.QUERY_REQUEST, 
				Mail.makeRequest(MessageId.REMOVE_SEARCH_WORD, new ArrayBox!(char)(word.word)));
		
		res_command.status = UIStatus.OK; // OK
		
		return res_command;
	}
	
	// 検索
	UIResponseCommand searchFile(UIRequestCommand req_command)
	{
		Mail				mail;
		UIRSearch	 		result;
		UIQSearch			search 	= new UIQSearch(req_command);
		
		debug(ui_task) mbswritef(derr, "UI: 検索(max: %d, key type: %d, sort type: %d, key:%s)\n",
			search.max, cast(uint)search.keyType, cast(uint)search.sortType, search.key);
		
		result	= searchFileUI(search);
		debug(ui_task) mbswritef(derr, "UI: 検索リターン\n");
		
		result.max = search.max;
		result.sort(search.sortType);
		
		result.status = UIStatus.OK; // OK
		
		search = null;
		
		return result;
	}
	
	// UI用検索 
	UIRSearch searchFileUI(UIQSearch search)
	{
		UIRSearch 			result = new UIRSearch;
		UIFileInformation[]	array;
		UIFileInformation	info;
		NyKeyInformation	keys[];
		
		switch (search.keyType) {
		case UISearchKeyType.KEYWORD:
			keys = environ.keyTable.searchName(search.key);
			break;
		case UISearchKeyType.TRIP:
			keys = environ.keyTable.searchTrip(search.key);
			break;
		case UISearchKeyType.HASH:
			ubyte[] hash = hexStringToBinary(search.key);
			keys = environ.keyTable.searchHash(hash);
			break;
		}
		
		foreach (NyKeyInformation e; keys) {
			info = new UIFileInformation;
			
			info.fileName 		= e.fileName.dup;
			info.fileSize		= e.fileSize;
			info.hash			= e.hash;
			info.trip			= e.trip;
			info.referredCount	= e.referredBlock;
			
			array ~= info;
		}
		
		result.record(array);
		
		debug(key_task) mbswritef(derr, "UI: キー検索(key:%s, hit:%d)\n",
													search.key, array.length);
		
		keys = null;
		array = null;
		info = null;
		
		return result;
	}
	
	// ノード情報取得
	UIResponseCommand getNodeInformation(UIRequestCommand req_command)
	{
		UIRGetNodeInformation	result;
		UIQGetNodeInformation	nodeinfo = new UIQGetNodeInformation(req_command);
		uint					bits = nodeinfo.nodeType;
		
		debug(ui_task) mbswritef(derr, "UI: ノード情報取得(type bits: %04x)\n", bits);
		
		result	= getNodeInformation(nodeinfo);
		
		result.status = UIStatus.OK; // OK
		
		return result;
	}
	
	// ノード情報取得
	UIRGetNodeInformation getNodeInformation(UIQGetNodeInformation nodeinfo)
	{
		UIRGetNodeInformation 		result	= new UIRGetNodeInformation;
		uint 						bits 	= nodeinfo.nodeType;
		UINodeDetails[]			nodes;
		
		if (bits & UINodeType.DOWNSTREAM_CONNECTION) {
			foreach (Connection e; environ.downstreamConnection.values(LockerMode.NON)) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				nodes ~= new UINodeDetails(
								UINodeType.DOWNSTREAM_CONNECTION, e.nodeDetails,
								e.connection.ctl, e.connection.rate
						);
			}
		}
		if (bits & UINodeType.UPSTREAM_CONNECTION) {
			foreach (Connection e; environ.upstreamConnection.values(LockerMode.NON)) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				nodes ~= new UINodeDetails(
								UINodeType.UPSTREAM_CONNECTION, e.nodeDetails,
								e.connection.ctl, e.connection.rate
						);
			}
		}
		if (bits & UINodeType.DATA_CONNECTION) {
			foreach (Connection e; environ.dataConnection.values(LockerMode.NON)) {
				if (e is null || e.connection is null || e.nodeDetails is null) {
					continue;
				}
				nodes ~= new UINodeDetails(
								UINodeType.DATA_CONNECTION, e.nodeDetails,
								e.connection.ctl, e.connection.rate
						);
			}
		}
		if (bits & UINodeType.UNCONNECTION) {
			foreach (NodeDetails e; environ.nodeTable.values(LockerMode.NON)) {
				if (e is null) {
					continue;
				}
				nodes ~= new UINodeDetails(UINodeType.UNCONNECTION, e);
			}
		}
		
		result.nodeDetails(nodes);
		result.status = UIStatus.OK;
		
		nodes = null;
		
		return result;
	}
	
	/*-------------------------------------------
	 *		タスク起動処理
	 *-------------------------------------------*/
	
	// 全タスク起動
	bool startTask()
	{
		// スレッドの優先順位を上げる
		//setPriority(PRIORITY.INCREASE);
		
		debug(ui_task)mbswritef(derr, "-- タスク起動中....\n");
		m_resource = new TaskResource;
		createResource(m_resource);
		
		environ.initialize();
		// 初期パラメータ設定
		config.fileSharingPort = m_ny_default_port;
		
		// スレッド生成
		
		m_node_task				= new NodeTask(environ, resource);
		m_key_task				= new KeyTask(environ, resource);
		m_query_task			= new QueryTask(environ, resource);
		m_control_task			= new ControlTask(environ, resource);
		m_cache_task			= new CacheTask(environ, resource);
		m_link_task				= new LinkTask(environ, resource);
		m_connection_task		= new ConnectionTask(environ, resource);
		
		/* 起動 */
		
		// キャッシュ
		startEvent(EventId.TASK_START, TaskId.CACHE);
		m_cache_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.CACHE, 10000)) {
			throw new Exception("--Cacheタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Cacheタスク起動 %08x\n", m_cache_task.id);
		
		// キー
		startEvent(EventId.TASK_START, TaskId.KEY);
		m_key_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.KEY, 10000)) {
			throw new Exception("--Keyタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Keyタスク起動 %08x\n", m_key_task.id);
		
		
		// ノード
		startEvent(EventId.TASK_START, TaskId.NODE);
		m_node_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.NODE, 10000)) {
			throw new Exception("--Nodeタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Nodeタスク起動 %08x\n", m_node_task.id);
		
		// クエリ
		startEvent(EventId.TASK_START, TaskId.QUERY);
		m_query_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.QUERY, 10000)) {
			throw new Exception("--Queryタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Queryタスク起動 %08x\n", m_query_task.id);
		
		// コントロール
		startEvent(EventId.TASK_START, TaskId.CONTROL);
		m_control_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.CONTROL, 10000)) {
			throw new Exception("--Controlタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Controlタスク起動 %08x\n", m_control_task.id);
		
		// リンク
		startEvent(EventId.TASK_START, TaskId.LINK);
		m_link_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.LINK, 10000)) {
			throw new Exception("--Linkタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Linkタスク起動 %08x\n", m_link_task.id);
		
		// コネクション
		startEvent(EventId.TASK_START, TaskId.CONNECTION);
		m_connection_task.start();
		if (!pollEndEvent(EventId.TASK_START, TaskId.CONNECTION, 10000)) {
			throw new Exception("--Connectionタスクの起動に失敗");
		}
		debug(ui_task)mbswritef(derr, "--Connectionタスク起動 %08x\n", m_connection_task.id);
		
		/* 全タスク起動通知 */
		startEvent(EventId.TASK_START, TaskId.ALL);
		debug(ui_task)mbswritef(derr, "-- 全タスク起動\n");
		
		// スレッドの優先順位を下げる
		//setPriority(PRIORITY.DECREASE);
		return true;
	}
	
	// 全タスク終了
	void endTask()
	{
		debug(ui_task)mbswritef(derr, "-- 全タスク停止中....\n");
		Mail end_message = Mail.makeRequest(MessageId.TERMINATE, null);
		
		sendRequestTo(MailId.KEY_REQUEST, end_message);
		sendRequestTo(MailId.QUERY_REQUEST, end_message);
		sendRequestTo(MailId.CONTROL_REQUEST, end_message);
		sendRequestTo(MailId.CACHE_REQUEST, end_message);
		sendRequestTo(MailId.LINK_REQUEST, end_message);
		sendRequestTo(MailId.NODE_REQUEST, end_message);
		sendRequestTo(MailId.CONNECTION_REQUEST, end_message);
		
		startEvent(EventId.TASK_START, TaskId.ALL);
//		pollEndEvent(EventId.TASK_START, 	TaskId.KEY|TaskId.QUERY|TaskId.CONTROL|
//											TaskId.CACHE|TaskId.LINK|TaskId.NODE|TaskId.CONNECTION, 5000);
		m_key_task.wait();
		m_query_task.wait();
		m_control_task.wait();
		m_cache_task.wait();
		m_link_task.wait();
		m_node_task.wait();
		m_connection_task.wait();
		
//		delete m_node_task;
//		delete m_key_task;
//		delete m_query_task;
//		delete m_control_task;
//		delete m_cache_task;
//		delete m_link_task;
//		delete m_connection_task;
		
		resource.mailBox.remove(); // 全て削除
		environ.terminate();
		
		debug(ui_task)mbswritef(derr, "-- 全タスク停止\n");
	}
	
	// 死んだとき用
	void VERY_UGLY_FUNCTION()
	{
		TaskId id;
		for (int i = 0; i < TaskId.sizeof * 8; ++i) {
			id = cast(TaskId)(1 << i);
			
			if (pollEndEvent(EventId.TASK_START, id)) {
				debug(ui_task) mbswritef(derr, "UI: タスクリブート(%08X)\n", cast(uint)id);
				endTask();
				Thread.getThis.pause();
				startTask();
			}
		}
		startEvent(EventId.TASK_START, TaskId.ALL);
	}
	
	/* タスクリソース作成	*/
	void createResource(TaskResource resource)
	{
		// メールボックス
		foreach (MailId id; g_mail_ids) {
			resource.mailBox.create(id);
		}
		
		// イベントフラグ
		foreach (EventId id; g_event_ids) {
			resource.eventFlag.create(id);
		}
		
		// セマフォ
		foreach (LockId id; g_lock_ids) {
			resource.semaphore.create(id);
		}
	}
	
public:
	this(uint uiport, uint ny_port)
	{
		super(MailId.UI_REQUEST, MailId.UI_RESPONSE, getEnvironment(), null);
		m_port				= uiport;
		m_ny_default_port 	= ny_port;
	}
}


